# -*- coding: utf-8 -*-

# Copyright (c) 2016-2023 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

import sys
import math

import numpy as np
import pandas as pd
import geojson
from collections.abc import Iterable

from pandapower.auxiliary import soft_dependency_error, version_check, MapboxTokenMissing
from pandapower.plotting.plotly.get_colors import get_plotly_color, get_plotly_cmap
from pandapower.plotting.plotly.mapbox_plot import _on_map_test, _get_mapbox_token

try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging
logger = logging.getLogger(__name__)

try:
    from plotly.graph_objs.scatter.marker import ColorBar
    from plotly.graph_objs import Figure, Layout
    from plotly.graph_objs.layout import XAxis, YAxis
    from plotly.graph_objs.scatter import Line, Marker
    from plotly.graph_objs.scattermapbox import Line as scmLine
    from plotly.graph_objs.scattermapbox import Marker as scmMarker

    version_check('plotly')
    PLOTLY_INSTALLED = True
except ImportError:
    PLOTLY_INSTALLED = False


def _in_ipynb():
    """
    An auxiliary function which checks if plot is called from a jupyter-notebook or not
    """
    try:
        from IPython import get_ipython
        return get_ipython().__class__.__name__ == 'ZMQInteractiveShell'
    except ImportError:
        return False


def sum_line_length(pts):
    pt_diff = lambda p: (p[0][0] - p[1][0], p[0][1] - p[1][1])
    diffs = map(pt_diff, zip(pts[:-1], pts[1:]))
    line_length = sum(math.hypot(d1, d2) for d1, d2 in diffs)
    return line_length


def get_line_neutral(coord):
    if len(coord) == 1:
        return coord[0]
    half_length = sum_line_length(coord) / 2.0
    length = 0.0
    ind = 0
    while length < half_length:
        ind += 1
        length = sum_line_length(coord[:ind])

    start_coord = coord[ind - 2]
    end_coord = coord[ind - 1]
    mid = [(a1 + a2) / 2.0 for a1, a2 in zip(start_coord, end_coord)]

    return mid


def create_edge_center_trace(line_trace, size=1, patch_type="circle", color="white", infofunc=None,
                             trace_name='edge_center', use_line_geo=False, showlegend=False,
                             legendgroup=None, hoverlabel=None):
    """
    Creates a plotly trace of pandapower buses.

    INPUT:
        **line traces** (from pandapowerNet) - The already generated line traces with center geodata

    OPTIONAL:

        **size** (int, 5) - patch size

        **patch_type** (str, "circle") - patch type, can be

                - "circle" for a circle
                - "square" for a rectangle
                - "diamond" for a diamond
                - much more patch types at https://plot.ly/python/reference/#scatter-marker

        **infofunc** (pd.Series, None) - hoverinfo for each trace element. Indices should correspond
        to the pandapower element indices

        **trace_name** (String, "buses") - name of the trace which will appear in the legend

        **color** (String, "blue") - color of buses in the trace

    """
    # color = get_plotly_color(color)

    center_trace = dict(type='scatter', text=[], mode='markers', hoverinfo='text', name=trace_name,
                        marker=dict(color=color, size=size, symbol=patch_type),
                        showlegend=showlegend, legendgroup=legendgroup)
    if hoverlabel is not None:
        center_trace.update({'hoverlabel': hoverlabel})

    if not use_line_geo:
        center_trace['x'], center_trace['y'] = (line_trace[0]["x"][1::4], line_trace[0]["y"][1::4])
    else:
        x, y = [], []
        for trace in line_trace:
            coord = list(zip(trace["x"], trace["y"]))
            mid_coord = get_line_neutral(coord)
            x.append(mid_coord[0])
            y.append(mid_coord[1])

        center_trace['x'], center_trace['y'] = (x, y)

    center_trace['text'] = infofunc

    return center_trace


def create_bus_trace(net, buses=None, size=5, patch_type="circle", color="blue", infofunc=None,
                     trace_name='buses', legendgroup=None, cmap=None, cmap_vals=None,
                     cbar_title=None, cmin=None, cmax=None, cpos=1.0, colormap_column="vm_pu"):
    """
    Creates a plotly trace of pandapower buses. It is a wrapper function for the more generic
    _create_node_trace function.

    INPUT:
        **net** (pandapowerNet) - The pandapower network

    OPTIONAL:
        **buses** (list, None) - The buses for which the collections are created.
        If None, all buses in the network are considered.

        **size** (int, 5) - patch size

        **patch_type** (str, "circle") - patch type, can be

                - "circle" for a circle
                - "square" for a rectangle
                - "diamond" for a diamond
                - much more patch types at https://plot.ly/python/reference/#scatter-marker

        **infofunc** (pd.Series, None) - hoverinfo for bus elements. Indices should correspond to
        the pandapower element indices

        **trace_name** (String, "buses") - name of the trace which will appear in the legend

        **color** (String, "blue") - color of buses in the trace

        **cmap** (String, None) - name of a colormap which exists within plotly
        (Greys, YlGnBu, Greens, YlOrRd, Bluered, RdBu, Reds, Blues, Picnic, Rainbow,
        Portland, Jet, Hot, Blackbody, Earth, Electric, Viridis) alternatively a custom
        discrete colormap can be used. Append "_r" for inversion.

        **cmap_vals** (list, None) - values used for coloring using colormap

        **cbar_title** (String, None) - title for the colorbar

        **cmin** (float, None) - colorbar range minimum

        **cmax** (float, None) - colorbar range maximum

        **cpos** (float, 1.1) - position of the colorbar

        **colormap_column** (str, "vm_pu") - set color of bus according to this variable

    """
    node_element = 'bus'
    branch_element = 'line'
    return _create_node_trace(net,
                              nodes=buses,
                              size=size,
                              patch_type=patch_type,
                              color=color,
                              infofunc=infofunc,
                              trace_name=trace_name,
                              legendgroup=legendgroup,
                              cmap=cmap,
                              cmap_vals=cmap_vals,
                              cbar_title=cbar_title,
                              cmin=cmin,
                              cmax=cmax,
                              cpos=cpos,
                              colormap_column=colormap_column,
                              node_element=node_element,
                              branch_element=branch_element)


def _create_node_trace(net, nodes=None, size=5, patch_type='circle', color='blue', infofunc=None,
                       trace_name='nodes', legendgroup=None, cmap=None, cmap_vals=None,
                       cbar_title=None, cmin=None, cmax=None, cpos=1.0, colormap_column='vm_pu',
                       node_element='bus', branch_element='line'):
    """
    Creates a plotly trace of node elements. In pandapower, it should be called by
    create_bus_traces. The rather generic, non-power net specific names were introduced to make it
    usable in other packages, e.g. for pipe networks.

    INPUT:
        **net** (pandapowerNet) - The network

    OPTIONAL:
        **nodes** (list, None) - The nodes for which the collections are created.
                                 If None, all nodes in the network are considered.

        **size** (int, 5) - patch size

        **patch_type** (str, "circle") - patch type, can be

                - "circle" for a circle
                - "square" for a rectangle
                - "diamond" for a diamond
                - much more patch types at https://plot.ly/python/reference/#scatter-marker

        **infofunc** (pd.Series, None) - hoverinfo for node elements. Indices should correspond to
                                         the node element indices

        **trace_name** (String, "buses") - name of the trace which will appear in the legend

        **color** (String, "blue") - color of nodes in the trace

        **cmap** (String, None) - name of a colormap which exists within plotly
            (Greys, YlGnBu, Greens, YlOrRd, Bluered, RdBu, Reds, Blues, Picnic, Rainbow,
            Portland, Jet, Hot, Blackbody, Earth, Electric, Viridis) alternatively a custom
            discrete colormap can be used

        **cmap_vals** (list, None) - values used for coloring using colormap

        **cbar_title** (String, None) - title for the colorbar

        **cmin** (float, None) - colorbar range minimum

        **cmax** (float, None) - colorbar range maximum

        **cpos** (float, 1.1) - position of the colorbar

        **colormap_column** (str, "vm_pu") - set color of bus according to this variable

        **node_element** (str, "bus") - name of the node element in the net. In a pandapower net,
                                        this is alwas "bus"

        **branch_element** (str, "line") - name of the branch element in the net. In a pandapower
                                           net, this is alwas "line"

    """
    if not PLOTLY_INSTALLED:
        soft_dependency_error(str(sys._getframe().f_code.co_name) + "()", "plotly")
    color = get_plotly_color(color)
    node_trace = dict(type='scatter', text=[], mode='markers', hoverinfo='text', name=trace_name,
                      marker=dict(color=color, size=size, symbol=patch_type))
    nodes = net[node_element].index.tolist() if nodes is None else list(nodes)
    node_plot_index = [b for b in nodes if b in list(set(nodes) & set(net[node_element]["geo"].index))]
    node_trace['x'], node_trace['y'] = zip(*net[node_element].loc[node_plot_index, 'geo'].dropna().apply(geojson.loads).apply(geojson.utils.coords).apply(next).to_list())
    if not isinstance(infofunc, pd.Series) and isinstance(infofunc, Iterable) and \
            len(infofunc) == len(nodes):
        infofunc = pd.Series(index=nodes, data=infofunc)
    node_trace['text'] = net[node_element].loc[node_plot_index, 'name'] if infofunc is None else \
        infofunc.loc[nodes]
    if legendgroup:
        node_trace['legendgroup'] = legendgroup
    # if color map is set
    if cmap is not None:
        # TODO introduce discrete colormaps (see contour plots in plotly)
        # if cmap_vals are not given

        cmap = 'Jet' if cmap is True else cmap

        if cmap_vals is not None:
            cmap_vals = cmap_vals
        else:
            if net["res_" + branch_element].shape[0] == 0:
                if branch_element == "line":
                    logger.error(
                        "There are no power flow results for buses voltage magnitudes which are"
                        "default for bus colormap coloring..."
                        "set cmap_vals input argument if you want colormap according to some "
                        "specific values...")
                else:
                    logger.error(
                        "There are no simulation results which are default for %s colormap coloring..."
                        "set cmap_vals input argument if you want colormap according to some "
                        "specific values..." % node_element)
            cmap_vals = net["res_" + node_element].loc[node_plot_index, colormap_column].values

        cmap_vals = net["res_" + node_element].loc[
            node_plot_index, colormap_column] if cmap_vals is None else cmap_vals

        cmin = cmap_vals.min() if cmin is None else cmin
        cmax = cmap_vals.max() if cmax is None else cmax

        node_trace['marker'] = Marker(size=size,
                                      color=cmap_vals, cmin=cmin, cmax=cmax,
                                      colorscale=cmap,
                                      colorbar=ColorBar(thickness=10,
                                                        x=cpos),
                                      symbol=patch_type
                                      )

        if cbar_title:
            node_trace['marker']['colorbar']['title'] = cbar_title

        node_trace['marker']['colorbar']['title']['side'] = 'right'
    return [node_trace]


def _get_branch_geodata_plotly(net, branches, use_branch_geodata, branch_element='line',
                               node_element='bus'):
    xs = []
    ys = []
    if use_branch_geodata:
        for line_ind, _ in branches.iterrows():
            line_coords = list(geojson.utils.coords(geojson.loads(net[branch_element].at[line_ind, 'geo'])))
            line_x, line_y = list(zip(*line_coords))
            xs += line_x
            xs += [None]
            ys += line_y
            ys += [None]
    else:
        # getting x and y values from bus_geodata for from and to side of each line
        n = node_element
        from_n = 'from_' + n
        to_n = 'to_' + n
        from_node_xy = net[n].loc[branches[from_n], 'geo'].apply(geojson.loads).apply(geojson.utils.coords).apply(next).tolist()
        to_node_xy = net[n].loc[branches[to_n], 'geo'].apply(geojson.loads).apply(geojson.utils.coords).apply(next).tolist()
        from_node_x, from_node_y = zip(*from_node_xy)
        to_node_x, to_node_y = zip(*to_node_xy)
        # center point added because of the hovertool
        center = (np.array(from_node_x) + np.array(to_node_x)) / 2
        none_list = [None] * len(from_node_x)
        xs = np.array([from_node_x, center, to_node_x, none_list]).T.flatten().tolist()

        # center point added because of the hovertool
        center = (np.array(from_node_y) + np.array(to_node_y)) / 2
        none_list = [None] * len(from_node_y)
        ys = np.array([from_node_y, center, to_node_y, none_list]).T.flatten().tolist()

    # [:-1] is because the trace will not appear on maps if None is at the end
    return xs[:-1], ys[:-1]


def create_line_trace(net, lines=None, use_line_geo=True, respect_switches=False, width=1.0,
                      color='grey', infofunc=None, trace_name='lines', legendgroup='lines',
                      cmap=None, cbar_title=None, show_colorbar=True, cmap_vals=None, cmin=None,
                      cmax=None, cpos=1.1, cmap_vals_category='loading_percent', hoverlabel=None,
                      dash="solid"):
    """
    Creates a plotly trace of pandapower lines. It is a power net specific wrapper function for the
    more generic _create_line_trace function.

    INPUT:
        **net** (pandapowerNet) - The pandapower network

    OPTIONAL:
        **lines** (list, None) - The lines for which the collections are created.
        If None, all lines in the network are considered.

        **width** (int, 1) - line width


        **infofunc** (pd.Series, None) - hoverinfo for line elements. Indices should correspond to
        the pandapower element indices

        **trace_name** (String, "lines") - name of the trace which will appear in the legend

        **color** (String, "grey") - color of lines in the trace

        **legendgroup** (String, None) - defines groups of layers that will be displayed in a legend
        e.g. groups according to voltage level (as used in `vlevel_plotly`)

        **cmap** (String, None) - name of a colormap which exists within plotly if set to True default `Jet`
        colormap is used, alternative colormaps : Greys, YlGnBu, Greens, YlOrRd,
        Bluered, RdBu, Reds, Blues, Picnic, Rainbow, Portland, Jet, Hot, Blackbody, Earth, Electric, Viridis
        To revert the color map, append "_r" to the name.

        **cmap_vals** (list, None) - values used for coloring using colormap

        **show_colorbar** (bool, False) - flag for showing or not corresponding colorbar

        **cbar_title** (String, None) - title for the colorbar

        **cmin** (float, None) - colorbar range minimum

        **cmax** (float, None) - colorbar range maximum

        **cpos** (float, 1.1) - position of the colorbar

        """

    branch_element = "line"
    node_element = "bus"
    separator_element = "switch"

    return _create_branch_trace(net=net,
                                branches=lines,
                                use_branch_geodata=use_line_geo,
                                respect_separators=respect_switches,
                                width=width,
                                color=color,
                                infofunc=infofunc,
                                trace_name=trace_name,
                                legendgroup=legendgroup,
                                cmap=cmap,
                                cbar_title=cbar_title,
                                show_colorbar=show_colorbar,
                                cmap_vals=cmap_vals,
                                cmin=cmin,
                                cmax=cmax,
                                cpos=cpos,
                                branch_element=branch_element,
                                separator_element=separator_element,
                                node_element=node_element,
                                cmap_vals_category=cmap_vals_category,
                                hoverlabel=hoverlabel,
                                dash=dash)

def create_dcline_trace(net, dclines=None, width=1.0,
                        color='grey', infofunc=None, trace_name='hvdc lines', legendgroup='dclines',
                        cmap=None, cbar_title=None, show_colorbar=True, cmap_vals=None, cmin=None,
                        cmax=None, cpos=1.1, cmap_vals_category='loading_percent', hoverlabel=None):
    """
    Creates a plotly trace of pandapower dclines. It is a power net specific wrapper function for the
    more generic _create_line_trace function.

    INPUT:
        **net** (pandapowerNet) - The pandapower network

    OPTIONAL:
        **dclines** (list, None) - The lines for which the collections are created.
        If None, all lines in the network are considered.

        **width** (int, 1) - line width

        **infofunc** (pd.Series, None) - hoverinfo for dcline elements. Indices should correspond to
        the pandapower element indices

        **trace_name** (String, "lines") - name of the trace which will appear in the legend

        **color** (String, "grey") - color of dclines in the trace

        **legendgroup** (String, None) - defines groups of layers that will be displayed in a legend
        e.g. groups according to voltage level (as used in `vlevel_plotly`)

        **cmap** (String, None) - name of a colormap which exists within plotly if set to True default `Jet`
        colormap is used, alternative colormaps : Greys, YlGnBu, Greens, YlOrRd,
        Bluered, RdBu, Reds, Blues, Picnic, Rainbow, Portland, Jet, Hot, Blackbody, Earth, Electric, Viridis
        To revert the color map, append "_r" to the name.

        **cmap_vals** (list, None) - values used for coloring using colormap

        **show_colorbar** (bool, False) - flag for showing or not corresponding colorbar

        **cbar_title** (String, None) - title for the colorbar

        **cmin** (float, None) - colorbar range minimum

        **cmax** (float, None) - colorbar range maximum

        **cpos** (float, 1.1) - position of the colorbar

        """

    branch_element = "dcline"
    node_element = "bus"
    separator_element = "switch"

    return _create_branch_trace(net=net,
                                branches=dclines,
                                use_branch_geodata=False,
                                width=width,
                                color=color,
                                infofunc=infofunc,
                                trace_name=trace_name,
                                legendgroup=legendgroup,
                                cmap=cmap,
                                cbar_title=cbar_title,
                                show_colorbar=show_colorbar,
                                cmap_vals=cmap_vals,
                                cmin=cmin,
                                cmax=cmax,
                                cpos=cpos,
                                branch_element=branch_element,
                                separator_element=separator_element,
                                node_element=node_element,
                                cmap_vals_category=cmap_vals_category,
                                hoverlabel=hoverlabel)


def _create_branch_trace(net, branches=None, use_branch_geodata=True, respect_separators=False,
                         width=1.0, color='grey', infofunc=None, trace_name='lines',
                         legendgroup=None, cmap=None, cbar_title=None, show_colorbar=True,
                         cmap_vals=None, cmin=None, cmax=None, cpos=1.1, branch_element='line',
                         separator_element='switch', node_element='bus',
                         cmap_vals_category='loading_percent', hoverlabel=None, dash="solid"):
    """
    Creates a plotly trace of branch elements. The rather generic, non-power net specific names
    were introduced to make it usable in other packages, e.g. for pipe networks.

   INPUT:
       **net** (pandapowerNet) - The network

   OPTIONAL:
       **branches** (list, None) - The branches for which the collections are created.
                                   If None, all branches in the network are considered.

       **use_branch_geodata** (bool, True) - whether the geodata of the branch tables should be used

       **respect_separators** (bool, True) - whether separating elements like switches should be
                                             considered

       **width** (int, 1) - branch width

       **color** (String, "grey") - color of lines in the trace

       **infofunc** (pd.Series, None) - hoverinfo for line elements. Indices should correspond to
           the pandapower element indices

       **trace_name** (String, "lines") - name of the trace which will appear in the legend

       **legendgroup** (String, None) - defines groups of layers that will be displayed in a legend
       e.g. groups according to voltage level (as used in `vlevel_plotly`)

       **cmap** (String, None) - name of a colormap which exists within plotly if set to True default `Jet`
       colormap is used, alternative colormaps : Greys, YlGnBu, Greens, YlOrRd,
       Bluered, RdBu, Reds, Blues, Picnic, Rainbow, Portland, Jet, Hot, Blackbody, Earth, Electric, Viridis

       **cmap_vals** (list, None) - values used for coloring using colormap

       **show_colorbar** (bool, False) - flag for showing or not corresponding colorbar

       **cbar_title** (String, None) - title for the colorbar

       **cmin** (float, None) - colorbar range minimum

       **cmax** (float, None) - colorbar range maximum

       **cpos** (float, 1.1) - position of the colorbar

       **branch_element** (str, "line") - name of the branch element in the net. In a pandapower
                                          net, this is alwas "line"

       **separator_element** (str, "switch") - name of the separator element in the net. In a
                                               pandapower net, this is alwas "switch"

      **node_element** (str, "bus") - name of the node element in the net. In a pandapower net,
                                      this is alwas "bus" (net.bus)

       """
    if not PLOTLY_INSTALLED:
        soft_dependency_error(str(sys._getframe().f_code.co_name) + "()", "plotly")

    color = get_plotly_color(color)

    # defining branches (lines) to be plot
    branches = net[branch_element].index.tolist() if branches is None else list(branches)
    if len(branches) == 0:
        return []

    if infofunc is not None:
        if not isinstance(infofunc, pd.Series) and isinstance(infofunc, Iterable) and \
                len(infofunc) == len(branches):
            infofunc = pd.Series(index=branches, data=infofunc)
        if len(infofunc) != len(branches) and len(infofunc) != len(net[branch_element]):
            raise UserWarning(f"Different amount of hover info than {branch_element}s to plot")
        assert isinstance(infofunc, pd.Series), \
            f"infofunc should be a pandas series with the net.{branch_element}.index " \
            f"to the infofunc contents"
    no_go_branches = set()
    if respect_separators:
        if separator_element == "switch":
            no_go_branches = set(branches) & \
                             set(net[separator_element].element[(net[separator_element].et == "l") &
                                                                (net[separator_element].closed == 0)])
        elif separator_element == "valve":
            no_go_branches = set(branches) & \
                             set(net[separator_element][(~net[separator_element].in_service) |
                                                        (net[separator_element].opened)])
        else:
            raise NotImplementedError(f"respect separators is only implements for switches, "
                                      "not for {separator_element}s.")
    branches_to_plot = net[branch_element].loc[list(set(net[branch_element].index) &
                                                    (set(branches) - no_go_branches))]
    no_go_branches_to_plot = None
    use_branch_geodata = use_branch_geodata if net[branch_element]["geo"].dropna().shape[0] > 0 else False
    if use_branch_geodata:
        branches_to_plot = branches_to_plot.loc[np.intersect1d(branches_to_plot.index,
                                                               net[branch_element].index)]
    else:
        branches_with_geodata = branches_to_plot['from_' + node_element].isin(
            net[node_element].dropna(subset=["geo"]).index) & \
                                branches_to_plot['to_' + node_element].isin(
                                    net[node_element].dropna(subset=["geo"]).index)
        branches_to_plot = branches_to_plot.loc[branches_with_geodata]
    cmap_branches = None
    if cmap is not None:
        # workaround: if colormap plot is used, each line need to be separate scatter object because
        # plotly still doesn't support appropriately colormap for line objects
        # TODO correct this when plotly solves existing github issue about Line colorbar

        cmap = 'jet' if cmap is True else cmap

        if cmap_vals is not None:
            if not isinstance(cmap_vals, np.ndarray):
                cmap_vals = np.asarray(cmap_vals)
        else:
            if net['res_' + branch_element].shape[0] == 0:
                logger.error(
                    "There are no simulation results for branches which are default for {}"
                    "colormap coloring..."
                    "set cmap_vals input argument if you want colormap according to some specific "
                    "values...".format(branch_element))
            cmap_vals = net['res_' + branch_element].loc[branches_to_plot.index,
            cmap_vals_category].values

        cmap_branches = get_plotly_cmap(cmap_vals, cmap_name=cmap, cmin=cmin, cmax=cmax)
        if len(cmap_branches) == len(net[branch_element]):
            # some branches are not plotted although cmap_value were provided for all branches
            branch_idx_map = dict(zip(net[branch_element].loc[branches].index.tolist(), range(len(branches))))
            cmap_branches = [cmap_branches[branch_idx_map[idx]] for idx in branches_to_plot.index]
        else:
            assert len(cmap_branches) == len(branches_to_plot), \
                "Different amounts of cmap values and branches to plot were supplied"
    branch_traces = []
    for col_i, (idx, branch) in enumerate(branches_to_plot.iterrows()):
        line_color = color
        line_info = branch['name'] if infofunc is None else infofunc.at[idx]
        if cmap is not None:
            try:
                line_color = cmap_branches[col_i]
            except IndexError:
                logger.warning("No color and info for {} {:d} (name: {}) available".format(
                    branch_element, idx, branch['name']))

        line_trace = dict(type='scatter',
                          text=[],
                          hoverinfo='text',
                          mode='lines',
                          name=trace_name,
                          line=Line(width=width, color=color, dash=dash),
                          showlegend=False,
                          legendgroup=legendgroup)

        line_trace['x'], line_trace['y'] = _get_branch_geodata_plotly(net,
                                                                      branches_to_plot.loc[idx:idx],
                                                                      use_branch_geodata,
                                                                      branch_element, node_element)

        line_trace['line']['color'] = line_color

        line_trace['text'] = line_info

        branch_traces.append(line_trace)

    # enable legend for one element to make the legend group show up in the legend:
    branch_traces[0]["showlegend"] = True

    if show_colorbar and cmap is not None:

        cmin = cmap_vals.min() if cmin is None else cmin
        cmax = cmap_vals.max() if cmax is None else cmax
        try:
            # TODO for custom colormaps
            cbar_cmap_name = 'Jet' if cmap == 'jet' else cmap
            # workaround to get colorbar for branches (an unvisible node is added)
            # get x and y of first line.from_bus:
            x, y = geojson.utils.coords(net[node_element].loc(net[branch_element][f"from_{node_element}"][net[branch_element].index[0]], "geo"))
            branches_cbar = dict(type='scatter', x=x, y=y, mode='markers',
                                 marker=Marker(size=0, cmin=cmin, cmax=cmax,
                                               color='rgb(255,255,255)',
                                               opacity=0,
                                               colorscale=cbar_cmap_name,
                                               colorbar=ColorBar(thickness=10,
                                                                 x=cpos),
                                               ))
            if cbar_title:
                branches_cbar['marker']['colorbar']['title'] = cbar_title

            branches_cbar['marker']['colorbar']['title']['side'] = 'right'

            branch_traces.append(branches_cbar)
        except:
            pass
    if len(no_go_branches) > 0:
        ng_traces = []
        no_go_branches_to_plot = net[branch_element].loc[list(no_go_branches)]
        for idx, branch in no_go_branches_to_plot.iterrows():
            line_color = color
            line_trace = dict(type='scatter',
                              text=[], hoverinfo='text', mode='lines', name='disconnected branches',
                              line=Line(width=width / 2, color='grey', dash='dot'),
                              legendgroup="disconnected " + legendgroup, showlegend=False)

            line_trace['x'], line_trace['y'] = _get_branch_geodata_plotly(net,
                                                                          no_go_branches_to_plot.loc[
                                                                          idx:idx],
                                                                          use_branch_geodata,
                                                                          branch_element, node_element)

            line_trace['line']['color'] = line_color
            try:
                line_trace['text'] = infofunc.loc[idx]
            except (KeyError, IndexError, AttributeError):
                line_trace["text"] = branch['name']

            ng_traces.append(line_trace)
        # enable legend for one element to make the legend group show up in the legend:
        ng_traces[0]["showlegend"] = True
        branch_traces += ng_traces

    # sort infofunc so that it is the correct order lines_to_plot + no_go_lines_to_plot
    if infofunc is not None:
        if not isinstance(infofunc, pd.Series) and isinstance(infofunc, Iterable) and \
                len(infofunc) == len(net[branch_element]):
            infofunc = pd.Series(index=net[branch_element].index, data=infofunc)
        assert isinstance(infofunc, pd.Series), \
            "infofunc should be a pandas series with the net.{}.index to the infofunc contents" \
                .format(branch_element)
        sorted_idx = branches_to_plot.index.tolist()
        if no_go_branches_to_plot is not None:
            sorted_idx += no_go_branches_to_plot.index.tolist()
        infofunc = infofunc.loc[sorted_idx]
    center_trace = create_edge_center_trace(branch_traces, color=color, infofunc=infofunc,
                                            use_line_geo=use_branch_geodata,
                                            showlegend=False, legendgroup=legendgroup,
                                            hoverlabel=hoverlabel)
    branch_traces.append(center_trace)

    return branch_traces


def create_trafo_trace(net, trafos=None, color='green', trafotype='2W', width=5, infofunc=None, cmap=None,
                       trace_name='2W transformers', cmin=None, cmax=None, cmap_vals=None, matching_params=None,
                       use_line_geo=None):
    """
    Creates a plotly trace of pandapower trafos.

    INPUT:
        **net** (pandapowerNet) - The pandapower network

    OPTIONAL:
        **trafos** (list, None) - The trafos for which the collections are created.
        If None, all trafos in the network are considered.

        **trafotype** (String, "2W") - trafotype can be 2W or 3W and can define which transformer table is taken (trafo or trafo3w)

        **width** (int, 5) - line width

        **infofunc** (pd.Series, None) - hoverinfo for trafo elements. Indices should correspond
        to the pandapower element indices

        **trace_name** (String, "lines") - name of the trace which will appear in the legend

        **color** (String, "green") - color of lines in the trace

        **cmap** (bool, False) - name of a colormap which exists within plotly (Greys, YlGnBu,
        Greens, YlOrRd, Bluered, RdBu, Reds, Blues, Picnic, Rainbow, Portland, Jet, Hot,
        Blackbody, Earth, Electric, Viridis)

        **cmap_vals** (list, None) - values used for coloring using colormap

        **cbar_title** (String, None) - title for the colorbar

        **cmin** (float, None) - colorbar range minimum

        **cmax** (float, None) - colorbar range maximum


    """
    if not PLOTLY_INSTALLED:
        soft_dependency_error(str(sys._getframe().f_code.co_name) + "()", "plotly")

    color = get_plotly_color(color)

    if trafotype == '2W':
        trafotable = 'trafo'

        trafo_buses_with_geodata = net.trafo.hv_bus.isin(net.bus.geo.dropna().index) & \
                                   net.trafo.lv_bus.isin(net.bus.geo.dropna().index)
        connections = [['hv_bus', 'lv_bus']]
    elif trafotype == '3W':
        trafotable = 'trafo3w'
        trafo_buses_with_geodata = net.trafo3w.hv_bus.isin(net.bus.dropna(subset=["geo"]).index) & \
                                   net.trafo3w.mv_bus.isin(net.bus.dropna(subset=["geo"]).index) & \
                                   net.trafo3w.lv_bus.isin(net.bus.dropna(subset=["geo"]).index)
        connections = [['hv_bus', 'lv_bus'], ['hv_bus', 'mv_bus'], ['mv_bus', 'lv_bus']]

    else:
        raise UserWarning('trafo_type should be 2W or 3W')
    # defining lines to be plot
    trafos = net[trafotable].index.tolist() if trafos is None else list(trafos)
    if len(trafos) == 0:
        return []

    trafos_mask = net[trafotable].index.isin(trafos)
    trafos_to_plot = net[trafotable].loc[trafo_buses_with_geodata & trafos_mask]

    if infofunc is not None:
        if not isinstance(infofunc, pd.Series) and isinstance(infofunc, Iterable) and \
                len(infofunc) == len(trafos):
            infofunc = pd.Series(index=trafos, data=infofunc)
        assert isinstance(infofunc, pd.Series), \
            "infofunc should be a pandas series with the net.trafo.index to the infofunc contents"
        infofunc = infofunc.loc[trafos_to_plot.index]

    cmap_colors = []
    if cmap is not None:
        cmap = 'jet' if cmap is None else cmap

        cmin = 0 if cmin is None else cmin
        cmax = 100 if cmin is None else cmax

        if cmap_vals is not None:
            cmap_vals = cmap_vals
        else:
            if net['res_' + trafotable].shape[0] == 0:
                logger.error("There are no power flow results for lines which are default for line colormap coloring..."
                             "set cmap_vals input argument if you want colormap according to some specific values...")
            cmap_vals = net['res_' + trafotable].loc[trafos_to_plot.index, 'loading_percent'].values

        cmap_colors = get_plotly_cmap(cmap_vals, cmap_name=cmap, cmin=cmin, cmax=cmax)

    trafo_traces = []

    for col_i, (idx, trafo) in enumerate(trafos_to_plot.iterrows()):
        if cmap is not None:
            color = cmap_colors[col_i]
        for from_bus1, to_bus1 in connections:

            trafo_trace = dict(type='scatter', text=[], line=Line(width=width, color=color),
                               hoverinfo='text', mode='lines', name=trace_name,
                               legendgroup=trace_name, showlegend=False)

            trafo_trace['text'] = trafo['name'] if infofunc is None else infofunc.loc[idx]

            from_bus = next(geojson.utils.coords(geojson.loads(net.bus.loc[trafo[from_bus1], "geo"])))
            to_bus = next(geojson.utils.coords(geojson.loads(net.bus.loc[trafo[to_bus1], "geo"])))
            trafo_trace["x"] = [from_bus[0], (from_bus[0] + to_bus[0]) / 2, to_bus[0]]
            trafo_trace["y"] = [from_bus[1], (from_bus[1] + to_bus[1]) / 2, to_bus[1]]

            trafo_traces.append(trafo_trace)
    trafo_traces[0]["showlegend"] = True

    center_trace = create_edge_center_trace(trafo_traces, color=color, infofunc=infofunc,
                                            use_line_geo=use_line_geo,
                                            showlegend=False, legendgroup=trace_name)
    trafo_traces.append(center_trace)
    return trafo_traces


def create_weighted_marker_trace(net, elm_type="load", elm_ids=None, column_to_plot="p_mw",
                                 sizemode="area", color="red", patch_type="circle",
                                 marker_scaling=1., trace_name="", infofunc=None,
                                 node_element="bus", show_scale_legend=True,
                                 scale_marker_size=None, scale_marker_color=None,
                                 scale_legend_unit=None, trace_kwargs=None):
    """Create a single-color plotly trace markers/patches (e.g., bubbles) of value-dependent size.

    Can be used with pandapipes.plotting.plotly.simple_plotly (pass as "additional_trace").
    If present in the respective pandapower-net table, the "in_service" and "scaling" column will be
    taken into account as factors to calculate the markers' weights.
    Negative values might lead to unexpected results, especially when pos. and neg. values are
    mixed in one column! All values are treated as absolute values.
    If value = 0, no marker will be created.

    INPUT:
        **net** (pandapowerNet) - the pandapower net of the plot

    OPTIONAL:
        **elm_type** (str, default "load") - the element table in the net that holds the values
        to plot

        **elm_ids** (list, default None) - the element IDs of the elm_type table for which markers
        will be created. If None, all IDs in net[elm_type] that are not 0 or NaN in the
        `column_to_plot` will be used.

        **column_to_plot** (str, default "p_mw") - the column in the net[elm_type] table that
        will be plotted. The suffix (everything behind the last '_' in the column name) will be
        used as the unit in the infofunction.

        **sizemode** (str, default "area") - whether the markers' "area" or "diameter"  will be
        proportional to the represented value

        **color** (str, default "red") - color for the markers

        **patch_type** (str, default "circle") - plotly marker style that will be used (other
        options are triangle-up, triangle-down and many more; for non-filled markers, append "-open"
        cf. https://plotly.com/python/marker-style/)

        **marker_scaling** (float, default 1.) - factor to scale the size of all markers

        **trace_name** (str, default ""): trace name for the legend. If empty, elm_type will be used.

        **infofunc** (pd.Series, default None): hover-infofuction to overwrite the internal infofunction

        **node_element** (str, default "bus") - the name of node elements in the net. "bus" for
        pandapower networks, "junction" for pandapipes networks

        **show_scale_legend** (bool, default True): display a marker legend at the top right of the
         plot

        **scale_marker_size** (float / list of floats, default None): adjust the size of the scale,
        gets multiplied with `marker_scaling`. Default size is the average size of the respective
        weighted markers rounded to 5. A list of values will result in several scale marker legends.

        **scale_marker_color** (str, default None): specify the color of the scale legend marker.
        If None, `color` will be used.

        **scale_legend_unit** (str, default None): specifies the unit shown in the scale legend
        marker's string. It does not trigger any unit conversions! If None, the last part of
        `column_to_plot` will be used (all upper case).

        **trace_kwargs** (dict, default None): additional/updated entries for the marker trace
        that is created in this function, e.g. {"visible": "legendonly"}

    OUTPUT:
        **marker_trace** (dict): dict for the plotly trace

    """
    # filter for relevant elements:
    elm_ids = net[elm_type].loc[~net[elm_type][column_to_plot].isna()].index.tolist() \
        if elm_ids is None else list(elm_ids)

    # apply the scaling and in_service column, if available
    elms_df = net[elm_type].loc[elm_ids]
    scaling = elms_df.scaling if "scaling" in elms_df.columns else 1
    in_service = elms_df.in_service if "in_service" in elms_df.columns else 1
    elms_df["values_scaled"] = elms_df[column_to_plot] * in_service * scaling

    # get sum per bus, if not a bus table already:
    if elm_type not in [node_element, "res_" + node_element]:
        values_by_bus = elms_df.groupby("bus").sum().values_scaled
    else:
        values_by_bus = elms_df.values_scaled
    values_by_bus = values_by_bus.loc[values_by_bus != 0]

    if any(values_by_bus < 0):
        logger.warning("A marker trace cannot be created for negative values!\n"
                       "They will be considered as absolute values. Items with negative values:\n"
                       + str(values_by_bus.loc[values_by_bus < 0]))

    # add geodata:
    xy_list = net[node_element].loc[values_by_bus.index, 'geo'].dropna().apply(geojson.loads).apply(geojson.utils.coords).apply(next)
    x_list, y_list = zip(*xy_list)

    # set up hover info:
    unit = column_to_plot.split("_")[-1]
    unit = unit.upper()
    trace_name = trace_name if len(trace_name) else elm_type
    if infofunc is None:
        infofunc = values_by_bus.apply(lambda x: f"{trace_name} {x:.2f} {unit}")
    if not isinstance(infofunc, pd.Series) and isinstance(infofunc, Iterable) and \
            len(infofunc) == len(values_by_bus.index):
        infofunc = pd.Series(index=values_by_bus.index, data=infofunc)

    # set up the marker trace:
    marker_trace = dict(type='scatter',
                        text=infofunc,
                        mode='markers',
                        hoverinfo='text',
                        name=trace_name,
                        x=list(x_list),
                        y=list(y_list))
    marker_trace["marker"] = dict(color=color,
                                  size=values_by_bus.abs() * marker_scaling,
                                  symbol=patch_type,
                                  sizemode=sizemode)

    # additional info for the create_scale_trace function:

    if not isinstance(scale_marker_size, Iterable):
        scale_marker_size = [scale_marker_size]

    marker_trace["meta"] = dict(marker_scaling=marker_scaling,
                                column_to_plot=column_to_plot,
                                scale_legend_unit=scale_legend_unit or
                                                  column_to_plot.split("_")[1].upper(),
                                show_scale_legend=show_scale_legend,
                                scale_marker_size=scale_marker_size,
                                scale_marker_color=scale_marker_color or color)
    if trace_kwargs is not None:
        marker_trace.update(trace_kwargs)
    return marker_trace


def create_scale_trace(net, weighted_trace, down_shift=0):
    """Create a scale (marker size legend) for a weighted_marker_trace.

    Will be used with pandapipes.plotting.plotly.simple_plotly, when "additional_trace" contains
    a trace created by :func:`create_weighted_marker_trace` with :code:`show_scale_legend=True`.
    The default reference marker is of average size of all weighted markers, rounded to the next 5,
    and comes with a string with the respective reference value and unit.

    INPUT:
        **net** (pandapowerNet) - the pandapower net of the plot

        **weighted_trace** (dict) - weighted plotly trace

        **down_shift** (int) - shift to align different scales below each other (prop. to y-offset)

    """
    marker = weighted_trace["marker"]
    scale_info = weighted_trace["meta"]
    unit = scale_info["scale_legend_unit"]  # p_mw...q_mvar ..

    # scale trace
    bus_x, bus_y = zip(*net.bus.geo.dropna().apply(geojson.loads).apply(geojson.utils.coords).apply(next))
    x_max, y_max = max(bus_x), max(bus_y)
    x_min, y_min = min(bus_x), min(bus_y)
    x_pos = x_max + (x_max - x_min) * 0.2
    x_pos2 = x_max + ((x_max - x_min) * 0.2 * 2)

    scale_traces = []
    for idx, sze in enumerate(scale_info["scale_marker_size"]):
        y_pos = y_max - ((y_max - y_min) * (0.2 * (down_shift + idx)))

        # default is the average rounded to 5
        if not sze:
            scale_size = math.ceil(marker["size"].mean() / 5) * 5
        else:
            scale_size = sze * scale_info['marker_scaling']

        # second (dummy) position is needed for correct marker sizing
        scale_trace = dict(type="scatter",
                           x=[x_pos, x_pos2],
                           y=[y_pos, y_pos],
                           mode="markers+text",
                           hoverinfo="skip",
                           marker=dict(size=[scale_size, 0],
                                       color=scale_info.get("scale_marker_color"),
                                       symbol=marker["symbol"],
                                       sizemode=marker["sizemode"]),
                           text=[f"{scale_size / scale_info['marker_scaling']:.0f} {unit}", ""],
                           textposition="top center",
                           showlegend=False,
                           textfont=dict(family="Helvetica",
                                         size=14,
                                         color="DarkSlateGrey"))
        scale_traces.append(scale_trace)

    return scale_traces


def draw_traces(traces, on_map=False, map_style='basic', showlegend=True, figsize=1,
                aspectratio='auto', filename='temp-plot.html', auto_open=True, **kwargs):
    """
    plots all the traces (which can be created using :func:`create_bus_trace`, :func:`create_line_trace`,
    :func:`create_trafo_trace`)
    to PLOTLY (see https://plot.ly/python/)

    INPUT:
        **traces** - list of dicts which correspond to plotly traces
        generated using: `create_bus_trace`, `create_line_trace`, `create_trafo_trace`

    OPTIONAL:
        **on_map** (bool, False) - enables using mapbox plot in plotly

        **map_style** (str, 'basic') - enables using mapbox plot in plotly

            - 'streets'
            - 'bright'
            - 'light'
            - 'dark'
            - 'satellite'

        **showlegend** (bool, 'True') - enables legend display

        **figsize** (float, 1) - aspectratio is multiplied by it in order to get final image size

        **aspectratio** (tuple, 'auto') - when 'auto' it preserves original aspect ratio of the
        network geodata any custom aspectration can be given as a tuple, e.g. (1.2, 1)

        **filename** (str, "temp-plot.html") - plots to a html file called filename. If None,
        no file will be created, just a plotly.Figure object will be returned

        **auto_open** (bool, 'True') - automatically open plot in browser

    OUTPUT:
        **figure** (graph_objs._figure.Figure) figure object

    """
    if not PLOTLY_INSTALLED:
        soft_dependency_error(str(sys._getframe().f_code.co_name) + "()", "plotly")
    if on_map:
        try:
            on_map = _on_map_test(traces[0]['x'][0], traces[0]['y'][0])
        except ImportError:
            logger.warning("Test if geo-data are in lat/long cannot be performed using geopy -> "
                           "plot errors are possible.")

        if on_map is False:
            logger.warning("Existing geodata are not real lat/lon geographical coordinates. -> "
                           "plot on maps is not possible.\n"
                           "Use geo_data_to_latlong(net, projection) to transform geodata from specific projection.")

    if on_map:
        # change traces for mapbox
        # change trace_type to scattermapbox and rename x to lat and y to lon
        for trace in traces:
            if 'x' in trace.keys():
                trace['lon'] = trace.pop('x')
            if 'y' in trace.keys():
                trace['lat'] = trace.pop('y')
            trace['type'] = 'scattermapbox'
            if "line" in trace and isinstance(trace["line"], Line):
                # scattermapboxplot lines do not support dash for some reason, make it a red line instead
                if "dash" in trace["line"]._props:
                    _prps = dict(trace["line"]._props)
                    _prps.pop("dash")
                    if trace["line"]._props["dash"] != "solid":
                        _prps["color"] = "red"
                    trace["line"] = scmLine(_prps)
                else:
                    trace["line"] = scmLine(dict(trace["line"]._props))
            elif "marker" in trace and isinstance(trace["marker"], Marker):
                trace["marker"] = scmMarker(trace["marker"]._props)

    # setting Figure object
    fig = Figure(data=traces,  # edge_trace
                 layout=Layout(
                     title_font=dict(size=16),
                     showlegend=showlegend,
                     autosize=(aspectratio == 'auto'),
                     hovermode='closest',
                     margin=dict(b=5, l=5, r=5, t=5),
                     # annotations=[dict(
                     #     text="",
                     #     showarrow=False,
                     #     xref="paper", yref="paper",
                     #     x=0.005, y=-0.002)],
                     xaxis=XAxis(showgrid=False, zeroline=False, showticklabels=False),
                     yaxis=YAxis(showgrid=False, zeroline=False, showticklabels=False),
                     # legend=dict(x=0, y=1.0)
                     legend={'itemsizing': 'constant'},
                 ),
                 )
    a = kwargs.get('annotation')
    if a:
        fig.add_annotation(a)

    # check if geodata are real geographical lat/lon coordinates using geopy
    if on_map:
        try:
            mapbox_access_token = _get_mapbox_token()
        except Exception:
            logger.exception('mapbox token required for map plots. '
                             'Get Mapbox token by signing in to https://www.mapbox.com/.\n'
                             'After getting a token, set it to pandapower using:\n'
                             'pandapower.plotting.plotly.mapbox_plot.set_mapbox_token(\'<token>\')')
            raise MapboxTokenMissing

        fig['layout']['mapbox'] = dict(accesstoken=mapbox_access_token,
                                       bearing=0,
                                       center=dict(lat=pd.Series(traces[0]['lat']).dropna().mean(),
                                                   lon=pd.Series(traces[0]['lon']).dropna().mean()),
                                       style=map_style,
                                       pitch=0,
                                       zoom=kwargs.pop('zoomlevel', 3))

    # default aspectratio: if on_map use auto, else use 'original'
    aspectratio = 'original' if not on_map and aspectratio == 'auto' else aspectratio

    if aspectratio != 'auto':
        if aspectratio == 'original':
            # TODO improve this workaround for getting original aspectratio
            xs = []
            ys = []
            for trace in traces:
                xs += trace.get('x') or trace.get('lon') or []
                ys += trace.get('y') or trace.get('lat') or []
            xs_arr = np.array(xs)
            ys_arr = np.array(ys)
            xrange = np.nanmax(xs_arr) - np.nanmin(xs_arr)
            yrange = np.nanmax(ys_arr) - np.nanmin(ys_arr)

            # the ratio only makes sense, if xrange and yrange != 0
            if xrange == 0 and yrange == 0:
                aspectratio = (1, 1)
            elif xrange == 0:
                aspectratio = (0.35, 1)
            elif yrange == 0:
                aspectratio = (1, 0.35)

            else:
                ratio = xrange / yrange
                if ratio < 1:
                    aspectratio = (ratio, 1.)
                else:
                    aspectratio = (1., 1 / ratio)

        aspectratio = np.array(aspectratio) / max(aspectratio)
        fig['layout']['width'], fig['layout']['height'] = ([ar * figsize * 700 for ar in aspectratio])

    # check if called from ipynb or not in order to consider appropriate plot function
    if _in_ipynb() and auto_open:
        from plotly.offline import init_notebook_mode, iplot as plot
        init_notebook_mode()
        plot(fig, filename=filename)
    elif filename is not None:
        from plotly.offline import plot as plot
        plot(fig, filename=filename, auto_open=auto_open)
    return fig



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2023 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.
from __future__ import annotations
import logging
import os
import re
import tempfile
import zipfile
from types import MappingProxyType
from typing import Dict, List
import pandas as pd
import numpy as np
from lxml import etree
from .other_classes import ReportContainer, Report, LogLevel, ReportCode
from .cim_tools import get_cim_schema


class CimParser:

    def __init__(self, cim: Dict[str, Dict[str, pd.DataFrame]] = None, cgmes_version: str = None, **kwargs):
        """
        This class parses CIM files and loads its content to a dictionary of
        CIM profile (dict) -> CIM element type (str) -> CIM elements (DataFrame)

        :param cim: CIM profile structure used for parsing
        :param cgmes_version: CIM version to use, '2.4.15' or '3.0', default '2.4.15'
        """
        self.logger = logging.getLogger(self.__class__.__name__)
        self.cgmes_version = '2.4.15' if cgmes_version is None else cgmes_version
        self.__cim_blueprint = self._initialize_cim_data_structure(self.cgmes_version)
        self.cim: Dict[str, Dict[str, pd.DataFrame]] = cim if cim is not None else self.get_cim_data_structure()
        self.file_names: Dict[str, str] = dict()
        self.report_container = ReportContainer()
        self.ignore_errors = bool(kwargs.get("ignore_errors", False))

    def parse_files(self, file_list: List[str] or str = None, encoding: str = None, prepare_cim_net: bool = False,
                    set_data_types: bool = False) -> CimParser:
        """
        Parse CIM XML files from a storage.

        :param file_list: The path to the CGMES files as a list. Note: The files need a FullModel to parse the
        CGMES profile. Optional, default: None.
        :param encoding: The encoding from the files. Optional, default: None
        :param prepare_cim_net: Set this parameter to True to prepare the parsed cim data according to the
        CimConverter. Optional, default: False
        :param set_data_types: Set this parameter to True to set the cim data types at the parsed data. Optional,
        default: False
        :return: Self
        """
        self.logger.info("Start parsing CIM files.")
        self.report_container.add_log(Report(level=LogLevel.INFO, code=ReportCode.INFO_PARSING,
                                             message="CIM parser starts parsing CIM files."))
        if file_list is not None:
            if isinstance(file_list, list):
                for file in file_list:
                    self._parse_source_file(file=file, output=self.cim, encoding=encoding)
            else:
                self._parse_source_file(file=file_list, output=self.cim, encoding=encoding)

        if prepare_cim_net:
            self.prepare_cim_net()
        if set_data_types:
            self.set_cim_data_types()
        self.logger.info("Finished parsing CIM files.")
        self.report_container.add_log(Report(level=LogLevel.INFO, code=ReportCode.INFO_PARSING,
                                             message="CIM parser finished parsing CIM files."))
        return self

    def set_cim_data_types(self) -> CimParser:
        """
        Set the data types from the columns from the DataFrames for each CIM element type and profile. Note: Currently
        only elements required for the CGMES converter are set.
        """
        self.logger.info("Setting the cim data types.")
        default_values = dict(
            {'positiveFlowIn': True, 'connected': True, 'length': 1., 'sections': 1, 'maximumSections': 1,
             'referencePriority': 999999, 'gch': 0., 'g0ch': 0.})  # todo check gch g0ch sections maximumSections
        to_bool = dict({'True': True, 'true': True, 'TRUE': True, True: True,
                        'False': False, 'false': False, 'FALSE': False, False: False,
                        'nan': False, 'NaN': False, 'NAN': False, 'Nan': False, np.nan: False})
        float_type = float
        int_type = pd.Int64Dtype()
        bool_type = pd.BooleanDtype()
        data_types_map = dict({'Float': float_type, 'Integer': int_type, 'Boolean': bool_type})
        cim_schema = get_cim_schema(self.cgmes_version)
        for profile in self.cim.keys():
            for cim_element_type, item in self.cim[profile].items():
                for col in item.columns:
                    # skip elements which are not available in the schema like FullModel
                    if cim_element_type not in cim_schema[profile]:
                        self.logger.debug("Skipping CIM element type %s from profile %s." % (cim_element_type, profile))
                        continue
                    if col in cim_schema[profile][cim_element_type]['fields'].keys() and \
                            'data_type_prim' in cim_schema[profile][cim_element_type]['fields'][col].keys():
                        data_type_col_str = cim_schema[profile][cim_element_type]['fields'][col]['data_type_prim']
                        if data_type_col_str in data_types_map.keys():
                            data_type_col = data_types_map[data_type_col_str]
                        else:
                            continue
                        self.logger.debug("Setting data type of %s from CIM element %s as type %s" %
                                          (col, cim_element_type, data_type_col_str))
                        if col in default_values.keys():  # todo deprecated due to repair function?
                            self.cim[profile][cim_element_type][col] = self.cim[profile][cim_element_type][col].fillna(
                                value=default_values[col])
                        if data_type_col == bool_type:
                            self.cim[profile][cim_element_type][col] = \
                                self.cim[profile][cim_element_type][col].map(to_bool)
                        try:
                            # special fix for integers:
                            if data_type_col == int_type:
                                self.cim[profile][cim_element_type][col] = \
                                    self.cim[profile][cim_element_type][col].astype(float_type)
                            self.cim[profile][cim_element_type][col] = \
                                self.cim[profile][cim_element_type][col].astype(data_type_col)
                        except Exception as e:
                            self.logger.warning("Couldn't set the datatype to %s for field %s at CIM type %s in "
                                                "profile %s!" % (data_type_col_str, col, cim_element_type, profile))
                            self.logger.warning("This may be harmless if the data is not need by the converter. "
                                                "Message: %s" % e)
        self.logger.info("Finished setting the cim data types.")
        self.report_container.add_log(Report(level=LogLevel.INFO, code=ReportCode.INFO_PARSING,
                                             message="CIM parser set the data types from the CIM data."))
        return self

    def prepare_cim_net(self) -> CimParser:
        """
        Make sure that the cim dictionaries only consists of valid DataFrames for each cim element type and append
        missing columns (not set but required CIM fields).
        """
        self.logger.info("Start preparing the cim data.")
        cim_data_structure = self.get_cim_data_structure()
        for profile in list(self.cim.keys()):
            if profile not in cim_data_structure.keys():
                # this profile is not used by the converter, drop it
                del self.cim[profile]
                continue
            for cim_element_type in list(self.cim[profile].keys()):
                # check if the CIM element type is a pd.DataFrame
                if not isinstance(self.cim[profile][cim_element_type], pd.DataFrame):
                    if profile in cim_data_structure.keys() and cim_element_type in cim_data_structure[profile].keys():
                        # replace the cim element type with the default empty DataFrame from the cim_data_structure
                        self.cim[profile][cim_element_type] = cim_data_structure[profile][cim_element_type]
                    else:
                        # this cim element type is not used by the converter, drop it
                        del self.cim[profile][cim_element_type]
                    self.logger.warning("%s isn't a DataFrame! The data won't be used!" % cim_element_type)

        # append missing columns to the CIM net
        for profile in cim_data_structure.keys():
            if profile not in self.cim.keys():
                self.cim[profile] = cim_data_structure[profile]
                continue
            for cim_element_type, item in cim_data_structure[profile].items():
                if cim_element_type not in self.cim[profile].keys():
                    self.cim[profile][cim_element_type] = cim_data_structure[profile][cim_element_type]
                    continue
                for column in item.columns:
                    if column not in self.cim[profile][cim_element_type].columns:
                        self.logger.info("Adding missing column %s to CIM element %s" % (column, cim_element_type))
                        self.cim[profile][cim_element_type][column] = np.nan

        # now remove columns which are not needed by the converter (to avoid renaming problems when merging DataFrames)
        for profile in cim_data_structure.keys():
            for cim_element_type in cim_data_structure[profile].keys():
                self.cim[profile][cim_element_type] = \
                    self.cim[profile][cim_element_type][cim_data_structure[profile][cim_element_type].columns]
        self.logger.info("Finished preparing the cim data.")
        self.report_container.add_log(Report(level=LogLevel.INFO, code=ReportCode.INFO_PARSING,
                                             message="CIM parser finished preparing the CIM data."))
        return self

    def _initialize_cim16_data_structure(self) -> Dict[str, Dict[str, pd.DataFrame]]:
        """
           Get the cim data structure used by the converter for cgmes version less than 3.
           :return Dict[str, Dict[str, pd.DataFrame]]: The cim data structure used by the converter.
           """
        self.logger.debug("Returning the CIM data structure.")
        return MappingProxyType({
            'eq': MappingProxyType({
                'ControlArea': pd.DataFrame(columns=['rdfId', 'name', 'type']),
                'TieFlow': pd.DataFrame(columns=['rdfId', 'Terminal', 'ControlArea', 'positiveFlowIn']),
                'ConnectivityNode': pd.DataFrame(columns=['rdfId', 'name', 'description', 'ConnectivityNodeContainer']),
                'Bay': pd.DataFrame(columns=['rdfId', 'VoltageLevel']),
                'BusbarSection': pd.DataFrame(columns=['rdfId', 'name']),
                'Substation': pd.DataFrame(columns=['rdfId', 'name', 'Region']),
                'GeographicalRegion': pd.DataFrame(columns=['rdfId', 'name']),
                'SubGeographicalRegion': pd.DataFrame(columns=['rdfId', 'name', 'Region']),
                'VoltageLevel': pd.DataFrame(columns=['rdfId', 'name', 'shortName', 'BaseVoltage', 'Substation']),
                'BaseVoltage': pd.DataFrame(columns=['rdfId', 'name', 'nominalVoltage']),
                'ExternalNetworkInjection': pd.DataFrame(columns=[
                    'rdfId', 'name', 'description', 'minP', 'maxP', 'minQ', 'maxQ', 'BaseVoltage', 'EquipmentContainer',
                    'RegulatingControl', 'governorSCD', 'maxInitialSymShCCurrent', 'minInitialSymShCCurrent',
                    'maxR1ToX1Ratio', 'minR1ToX1Ratio', 'maxR0ToX0Ratio', 'maxZ0ToZ1Ratio']),
                'ACLineSegment': pd.DataFrame(columns=[
                    'rdfId', 'name', 'description', 'length', 'r', 'x', 'bch', 'gch', 'r0', 'x0', 'b0ch', 'g0ch',
                    'shortCircuitEndTemperature', 'BaseVoltage', 'EquipmentContainer']),
                'Terminal': pd.DataFrame(columns=[
                    'rdfId', 'name', 'ConnectivityNode', 'ConductingEquipment', 'sequenceNumber']),
                'OperationalLimitSet': pd.DataFrame(columns=['rdfId', 'name', 'Terminal']),
                'OperationalLimitType': pd.DataFrame(columns=['rdfId', 'name', 'limitType']),
                'CurrentLimit': pd.DataFrame(columns=[
                    'rdfId', 'name', 'OperationalLimitSet', 'OperationalLimitType', 'value']),
                'VoltageLimit': pd.DataFrame(columns=[
                    'rdfId', 'name', 'OperationalLimitSet', 'OperationalLimitType', 'value']),
                'DCNode': pd.DataFrame(columns=['rdfId', 'name', 'DCEquipmentContainer']),
                'DCEquipmentContainer': pd.DataFrame(columns=['rdfId', 'name']),
                'DCLine': pd.DataFrame(columns=['rdfId', 'name']),
                'DCConverterUnit': pd.DataFrame(columns=['rdfId', 'name', 'Substation', 'operationMode']),
                'DCLineSegment': pd.DataFrame(columns=['rdfId', 'name', 'description', 'EquipmentContainer']),
                'CsConverter': pd.DataFrame(columns=['rdfId', 'BaseVoltage', 'ratedUdc']),
                'VsConverter': pd.DataFrame(columns=['rdfId', 'name', 'BaseVoltage', 'EquipmentContainer', 'ratedUdc']),
                'DCTerminal': pd.DataFrame(columns=[
                    'rdfId', 'name', 'DCNode', 'DCConductingEquipment', 'sequenceNumber']),
                'ACDCConverterDCTerminal': pd.DataFrame(columns=[
                    'rdfId', 'name', 'DCNode', 'DCConductingEquipment', 'sequenceNumber']),
                'Breaker': pd.DataFrame(columns=[
                    'rdfId', 'name', 'description', 'EquipmentContainer', 'normalOpen', 'retained']),
                'Disconnector': pd.DataFrame(columns=[
                    'rdfId', 'name', 'description', 'EquipmentContainer', 'normalOpen', 'retained']),
                'Switch': pd.DataFrame(columns=[
                    'rdfId', 'name', 'description', 'EquipmentContainer', 'normalOpen', 'retained']),
                'LoadBreakSwitch': pd.DataFrame(columns=[
                    'rdfId', 'name', 'description', 'EquipmentContainer', 'normalOpen', 'retained']),
                'EnergyConsumer': pd.DataFrame(columns=[
                    'rdfId', 'name', 'description', 'BaseVoltage', 'EquipmentContainer']),
                'ConformLoad': pd.DataFrame(columns=['rdfId', 'name', 'description']),
                'NonConformLoad': pd.DataFrame(columns=['rdfId', 'name', 'description']),
                'StationSupply': pd.DataFrame(columns=['rdfId', 'name', 'description', 'BaseVoltage']),
                'GeneratingUnit': pd.DataFrame(columns=[
                    'rdfId', 'name', 'nominalP', 'initialP', 'minOperatingP', 'maxOperatingP', 'EquipmentContainer', 'governorSCD']),
                'WindGeneratingUnit': pd.DataFrame(columns=['rdfId', 'nominalP', 'minOperatingP', 'maxOperatingP', 'governorSCD']),
                'HydroGeneratingUnit': pd.DataFrame(columns=['rdfId', 'nominalP', 'minOperatingP', 'maxOperatingP', 'governorSCD']),
                'SolarGeneratingUnit': pd.DataFrame(columns=['rdfId', 'nominalP', 'minOperatingP', 'maxOperatingP', 'governorSCD']),
                'ThermalGeneratingUnit': pd.DataFrame(columns=['rdfId', 'nominalP', 'minOperatingP', 'maxOperatingP', 'governorSCD']),
                'NuclearGeneratingUnit': pd.DataFrame(columns=['rdfId', 'nominalP', 'minOperatingP', 'maxOperatingP', 'governorSCD']),
                'RegulatingControl': pd.DataFrame(columns=['rdfId', 'name', 'mode', 'Terminal']),
                'SynchronousMachine': pd.DataFrame(columns=[
                    'rdfId', 'name', 'description', 'GeneratingUnit', 'EquipmentContainer', 'ratedU', 'ratedS', 'type', 'InitialReactiveCapabilityCurve',
                    'r2', 'x2', 'ratedPowerFactor', 'voltageRegulationRange', 'minQ', 'maxQ', 'RegulatingControl']),
                'AsynchronousMachine': pd.DataFrame(columns=[
                    'rdfId', 'name', 'description', 'GeneratingUnit', 'ratedS', 'ratedU', 'ratedPowerFactor',
                    'rxLockedRotorRatio', 'iaIrRatio', 'efficiency', 'ratedMechanicalPower']),
                'EnergySource': pd.DataFrame(columns=[
                    'rdfId', 'name', 'description', 'nominalVoltage', 'EnergySchedulingType', 'BaseVoltage',
                    'EquipmentContainer', 'voltageAngle', 'voltageMagnitude']),
                'EnergySchedulingType': pd.DataFrame(columns=['rdfId', 'name']),
                'StaticVarCompensator': pd.DataFrame(columns=['rdfId', 'name', 'description', 'voltageSetPoint','sVCControlMode']),
                'PowerTransformer': pd.DataFrame(columns=[
                    'rdfId', 'name', 'description', 'EquipmentContainer', 'isPartOfGeneratorUnit']),
                'PowerTransformerEnd': pd.DataFrame(columns=[
                    'rdfId', 'name', 'PowerTransformer', 'endNumber', 'Terminal', 'ratedS', 'ratedU',
                    'r', 'x', 'r0', 'x0', 'b', 'g', 'BaseVoltage', 'phaseAngleClock', 'connectionKind', 'grounded',
                    'xground']),
                'TapChangerControl': pd.DataFrame(columns=['rdfId', 'name', 'mode', 'Terminal']),
                'RatioTapChanger': pd.DataFrame(columns=[
                    'rdfId', 'name', 'TransformerEnd', 'neutralStep', 'lowStep', 'highStep', 'stepVoltageIncrement',
                    'neutralU', 'normalStep', 'ltcFlag', 'tculControlMode', 'TapChangerControl',
                    'RatioTapChangerTable']),
                'PhaseTapChangerLinear': pd.DataFrame(columns=[
                    'rdfId', 'TransformerEnd', 'neutralStep', 'lowStep', 'highStep', 'stepPhaseShiftIncrement',
                    'TapChangerControl']),
                'PhaseTapChangerAsymmetrical': pd.DataFrame(columns=[
                    'rdfId', 'TransformerEnd', 'neutralStep', 'lowStep', 'highStep', 'voltageStepIncrement',
                    'TapChangerControl']),
                'PhaseTapChangerSymmetrical': pd.DataFrame(columns=[
                    'rdfId', 'TransformerEnd', 'neutralStep', 'lowStep', 'highStep', 'voltageStepIncrement',
                    'TapChangerControl']),
                'PhaseTapChangerTabular': pd.DataFrame(columns=[
                    'rdfId', 'TransformerEnd', 'PhaseTapChangerTable', 'highStep', 'lowStep', 'neutralStep',
                    'TapChangerControl']),
                'PhaseTapChangerTablePoint': pd.DataFrame(columns=[
                    'rdfId', 'PhaseTapChangerTable', 'step', 'angle', 'ratio', 'r', 'x']),
                'RatioTapChangerTable': pd.DataFrame(columns=['rdfId', 'TransformerEnd', 'RatioTapChangerTable',
                                                              'highStep', 'lowStep', 'neutralStep']),
                'RatioTapChangerTablePoint': pd.DataFrame(columns=['rdfId', 'RatioTapChangerTable', 'step',
                                                                   'r', 'x', 'ratio']),
                'LinearShuntCompensator': pd.DataFrame(columns=[
                    'rdfId', 'name', 'description', 'nomU', 'gPerSection', 'bPerSection', 'maximumSections',
                    'normalSections', 'EquipmentContainer']),
                'NonlinearShuntCompensator': pd.DataFrame(columns=[
                    'rdfId', 'name', 'description', 'nomU', 'maximumSections']),
                'NonlinearShuntCompensatorPoint': pd.DataFrame(columns=[
                    'rdfId', 'description', 'NonlinearShuntCompensator', 'sectionNumber', 'b', 'g']),
                'EquivalentBranch': pd.DataFrame(columns=[
                    'rdfId', 'name', 'description', 'BaseVoltage', 'r', 'x', 'r21', 'x21', 'zeroR12', 'zeroR21',
                    'zeroX12', 'zeroX21']),
                'EquivalentInjection': pd.DataFrame(columns=['rdfId', 'name', 'description', 'BaseVoltage', 'r', 'x',
                                                             'regulationCapability']),
                'SeriesCompensator': pd.DataFrame(columns=[
                    'rdfId', 'name', 'description', 'BaseVoltage', 'r', 'x', 'r0', 'x0']),
                'Analog': pd.DataFrame(columns=[
                    'rdfId', 'name', 'description', 'measurementType', 'unitSymbol', 'unitMultiplier', 'Terminal',
                    'PowerSystemResource', 'positiveFlowIn']),
                'AnalogValue': pd.DataFrame(columns=[
                    'rdfId', 'name', 'sensorAccuracy', 'MeasurementValueSource', 'Analog', 'value']),
                'MeasurementValueSource': pd.DataFrame(columns=['rdfId', 'name']),
                'ReactiveCapabilityCurve': pd.DataFrame(columns=['rdfId', 'name', 'curveStyle', 'xUnit', 'y1Unit',
                                                                 'y2Unit']),
                'CurveData': pd.DataFrame(columns=['rdfId', 'Curve', 'xvalue', 'y1value', 'y2value'])

            }),
            'eq_bd': MappingProxyType({
                'ConnectivityNode': pd.DataFrame(columns=['rdfId', 'name', 'ConnectivityNodeContainer']),
                'BaseVoltage': pd.DataFrame(columns=['rdfId', 'name', 'nominalVoltage']),
                'Terminal': pd.DataFrame(
                    columns=['rdfId', 'ConnectivityNode', 'ConductingEquipment', 'sequenceNumber']),
                'EnergySource': pd.DataFrame(columns=['rdfId', 'name', 'nominalVoltage', 'EnergySchedulingType']),
                'EnergySchedulingType': pd.DataFrame(columns=['rdfId', 'name'])
            }),
            'ssh': MappingProxyType({
                'ControlArea': pd.DataFrame(columns=['rdfId', 'netInterchange']),
                'ExternalNetworkInjection': pd.DataFrame(columns=[
                    'rdfId', 'p', 'q', 'referencePriority', 'controlEnabled']),
                'Terminal': pd.DataFrame(columns=['rdfId', 'connected']),
                'DCTerminal': pd.DataFrame(columns=['rdfId', 'connected']),
                'ACDCConverterDCTerminal': pd.DataFrame(columns=['rdfId', 'connected']),
                'CsConverter': pd.DataFrame(columns=['rdfId', 'p', 'q']),
                'VsConverter': pd.DataFrame(columns=[
                    'rdfId', 'p', 'q', 'targetUpcc', 'droop', 'droopCompensation', 'qShare', 'targetUdc', 'targetPpcc',
                    'targetQpcc', 'pPccControl', 'qPccControl']),
                'Breaker': pd.DataFrame(columns=['rdfId', 'open']),
                'Disconnector': pd.DataFrame(columns=['rdfId', 'open']),
                'Switch': pd.DataFrame(columns=['rdfId', 'open']),
                'LoadBreakSwitch': pd.DataFrame(columns=['rdfId', 'open']),
                'EnergyConsumer': pd.DataFrame(columns=['rdfId', 'p', 'q']),
                'ConformLoad': pd.DataFrame(columns=['rdfId', 'p', 'q']),
                'NonConformLoad': pd.DataFrame(columns=['rdfId', 'p', 'q']),
                'StationSupply': pd.DataFrame(columns=['rdfId', 'p', 'q']),
                'RegulatingControl': pd.DataFrame(columns=[
                    'rdfId', 'discrete', 'enabled', 'targetValue', 'targetValueUnitMultiplier']),
                'SynchronousMachine': pd.DataFrame(columns=[
                    'rdfId', 'p', 'q', 'referencePriority', 'operatingMode', 'controlEnabled']),
                'AsynchronousMachine': pd.DataFrame(columns=['rdfId', 'p', 'q']),
                'EnergySource': pd.DataFrame(columns=['rdfId', 'activePower', 'reactivePower']),
                'StaticVarCompensator': pd.DataFrame(columns=['rdfId', 'q']),
                'TapChangerControl': pd.DataFrame(columns=[
                    'rdfId', 'discrete', 'enabled', 'targetValue', 'targetValueUnitMultiplier', 'targetDeadband']),
                'RatioTapChanger': pd.DataFrame(columns=['rdfId', 'step', 'controlEnabled']),
                'PhaseTapChangerLinear': pd.DataFrame(columns=['rdfId', 'step']),
                'PhaseTapChangerAsymmetrical': pd.DataFrame(columns=['rdfId', 'step']),
                'PhaseTapChangerSymmetrical': pd.DataFrame(columns=['rdfId', 'step']),
                'PhaseTapChangerTabular': pd.DataFrame(columns=['rdfId', 'step']),
                'LinearShuntCompensator': pd.DataFrame(columns=['rdfId', 'controlEnabled', 'sections']),
                'NonlinearShuntCompensator': pd.DataFrame(columns=['rdfId', 'controlEnabled', 'sections']),
                'EquivalentInjection': pd.DataFrame(columns=['rdfId', 'regulationTarget', 'regulationStatus', 'p', 'q'])
            }),
            'sv': MappingProxyType({
                'SvVoltage': pd.DataFrame(columns=['rdfId', 'TopologicalNode', 'v', 'angle']),
                'SvPowerFlow': pd.DataFrame(columns=['rdfId', 'Terminal', 'p', 'q']),
                'SvShuntCompensatorSections': pd.DataFrame(columns=['rdfId', 'ShuntCompensator', 'sections']),
                'SvTapStep': pd.DataFrame(columns=['rdfId', 'TapChanger', 'position'])
            }),
            'tp': MappingProxyType({
                'TopologicalNode': pd.DataFrame(columns=[
                    'rdfId', 'name', 'description', 'ConnectivityNodeContainer', 'BaseVoltage']),
                'DCTopologicalNode': pd.DataFrame(columns=['rdfId', 'name', 'DCEquipmentContainer']),
                'ConnectivityNode': pd.DataFrame(columns=['rdfId', 'TopologicalNode']),
                'Terminal': pd.DataFrame(columns=['rdfId', 'TopologicalNode']),
                'DCTerminal': pd.DataFrame(columns=['rdfId', 'DCTopologicalNode']),
                'ACDCConverterDCTerminal': pd.DataFrame(columns=['rdfId', 'DCTopologicalNode'])
            }),
            'tp_bd': MappingProxyType({
                'TopologicalNode': pd.DataFrame(columns=['rdfId', 'name', 'ConnectivityNodeContainer', 'BaseVoltage']),
                'ConnectivityNode': pd.DataFrame(columns=['rdfId', 'TopologicalNode'])
            }),
            'dl': MappingProxyType({
                'Diagram': pd.DataFrame(columns=['rdfId', 'name']),
                'DiagramObject': pd.DataFrame(columns=['rdfId', 'IdentifiedObject', 'Diagram', 'name']),
                'DiagramObjectPoint': pd.DataFrame(columns=[
                    'rdfId', 'sequenceNumber', 'xPosition', 'yPosition', 'DiagramObject'])}),
            'gl': MappingProxyType({
                'CoordinateSystem': pd.DataFrame(columns=['rdfId', 'name', 'crsUrn']),
                'Location': pd.DataFrame(columns=['rdfId', 'PowerSystemResources', 'CoordinateSystem']),
                'PositionPoint': pd.DataFrame(columns=['rdfId', 'Location', 'sequenceNumber', 'xPosition', 'yPosition'])
            })})

    def _parse_element(self, element, parsed=None):
        if parsed is None:
            parsed = dict()
        for key in element.keys():
            combined_key = element.tag + '-' + key
            if combined_key not in parsed:
                parsed[combined_key] = element.attrib.get(key)
            else:
                if not isinstance(parsed[combined_key], list):
                    parsed[combined_key] = [parsed[combined_key]]
                parsed[combined_key].append(element.attrib.get(key))
        if element.tag not in parsed and element.text is not None and element.text.strip(' \t\n\r'):
            parsed[element.tag] = element.text
        for child in list(element):
            self._parse_element(child, parsed)
        return parsed

    def _get_df(self, items):
        return pd.DataFrame([self._parse_element(child) for child in iter(items)])

    def _get_cgmes_profile_from_xml(self, root: etree.Element, default_profile: str = 'unknown') -> str:
        """
        Get the CGMES profile from the XML file.

        :param root: The root element from the XML tree
        :param default_profile: The default profile name which will be returned if ignore_errors is set to True.
        Optional, default: 'unknown'
        :return: The profile in short from: 'eq' for Equipment, 'eq_bd' for EquipmentBoundary,
        'ssh' for SteadyStateHypothesis, 'sv' for StateVariables,
        'tp' for Topology, 'tp_bd' for TopologyBoundary
        """
        element_types = pd.Series([ele.tag for ele in list(root)])
        element_types = element_types.drop_duplicates()
        full_model = element_types.str.find('FullModel')
        if full_model.max() >= 0:
            full_model = element_types[full_model >= 0].values[0]
        else:
            full_model = 'FullModel'
        full_model_profile = full_model[:-9] + 'Model.profile'
        full_model_df = self._get_df(root.findall('.//' + full_model))
        if full_model_df.index.size == 0 and self.ignore_errors:
            self.logger.warning("The FullModel is not given in the XML tree, returning %s" % default_profile)
            return default_profile
        elif full_model_df.index.size == 0:
            raise Exception("The FullModel is not given in the XML tree.")
        if full_model_df.index.size > 1 and self.ignore_errors:
            self.logger.warning("It is more than one FullModel given, returning the profile from the first FullModel.")
        elif full_model_df.index.size > 1:
            raise Exception("It is more than one FullModel given.")
        if full_model_profile not in full_model_df.columns and self.ignore_errors:
            self.logger.warning("The profile is not given in the FullModel, returning %s" % default_profile)
            return default_profile
        elif full_model_profile not in full_model_df.columns:
            raise Exception("The profile is not given in the FullModel.")
        profile_list = full_model_df[full_model_profile].values[0]
        if not isinstance(profile_list, list):
            profile_list = [profile_list]
        for one_profile in profile_list:
            if '/EquipmentCore/' in one_profile or '/EquipmentOperation/' in one_profile or \
                    '/EquipmentShortCircuit/' in one_profile or '/CoreEquipment-EU/' in one_profile:
                return 'eq'
            elif '/SteadyStateHypothesis/' in one_profile or '/SteadyStateHypothesis-EU/' in one_profile:
                return 'ssh'
            elif '/StateVariables/' in one_profile or '/StateVariables-EU/' in one_profile:
                return 'sv'
            elif '/Topology/' in one_profile or '/Topology-EU/' in one_profile:
                return 'tp'
            elif '/DiagramLayout/' in one_profile or '/DiagramLayout-EU/' in one_profile:
                return 'dl'
            elif '/GeographicalLocation/' in one_profile or '/GeographicalLocation-EU/' in one_profile:
                return 'gl'
            elif ('/EquipmentBoundary/' in one_profile or '/EquipmentBoundaryOperation/' in one_profile
                  or '/EquipmentBoundary-EU/' in one_profile):
                return 'eq_bd'
            elif '/TopologyBoundary/' in one_profile:
                return 'tp_bd'
            elif '/Operation-EU/' in one_profile:
                return 'op'
            elif '/ShortCircuit-EU/' in one_profile:
                return 'sc'
            elif '/Dynamics/' in one_profile:
                return 'dy'
        if self.ignore_errors:
            self.logger.warning("The CGMES profile could not be parsed from the XML, returning %s" % default_profile)
            self.report_container.add_log(Report(level=LogLevel.ERROR, code=ReportCode.ERROR_PARSING,
                                                 message="The CGMES profile could not be parsed from the XML, "
                                                         "returning %s" % default_profile))
            return default_profile
        else:
            self.report_container.add_log(Report(level=LogLevel.ERROR, code=ReportCode.ERROR_PARSING,
                                                 message="The CGMES profile could not be parsed from the XML."))
            raise Exception("The CGMES profile could not be parsed from the XML.")

    def _parse_source_file(self, file: str, output: dict, encoding: str, profile_name: str = None):
        self.logger.info("Parsing file: %s" % file)
        if not self._check_file(file):
            return
        # check if the file is a zip archive
        if file.lower().endswith('.zip'):
            # extract the zip in a temporary folder and delete it later
            temp_dir = tempfile.TemporaryDirectory()
            temp_dir_path = os.path.realpath(temp_dir.name)
            with zipfile.ZipFile(file, "r") as zip_ref:
                zip_ref.extractall(temp_dir_path)
            # parse the extracted CIM files
            for temp_file in os.listdir(temp_dir_path):
                temp_file = os.path.join(temp_dir_path, temp_file)
                if os.path.isfile(temp_file):
                    self._parse_source_file(temp_file, output=output, encoding=encoding)
                elif os.path.isdir(temp_file):
                    for sub_temp_file in os.listdir(temp_file):
                        sub_temp_file = os.path.join(temp_file, sub_temp_file)
                        self._parse_source_file(sub_temp_file, output=output, encoding=encoding)
            temp_dir.cleanup()
            del temp_dir, temp_dir_path
            return
        parser = etree.XMLParser(encoding=encoding, resolve_entities=False)
        xml_tree = etree.parse(file, parser)
        if profile_name is None:
            prf = self._get_cgmes_profile_from_xml(xml_tree.getroot())
        else:
            prf = profile_name
        self.file_names[prf] = file
        self._parse_xml_tree(xml_tree.getroot(), prf, output)

    def _parse_xml_tree(self, xml_tree: etree.ElementTree, profile_name: str, output: Dict | None = None):
        output = self.cim if output is None else output
        # get all CIM elements to parse
        element_types = pd.Series([ele.tag for ele in list(xml_tree)])
        element_types = element_types.drop_duplicates()
        prf_content: Dict[str, pd.DataFrame] = dict()
        ns_dict = dict()
        prf = profile_name
        if prf not in ns_dict.keys():
            ns_dict[prf] = dict()
        for _, element_type in element_types.items():
            if not isinstance(element_type, str):
                continue
            element_type_c = re.sub('{.*}', '', element_type)
            prf_content[element_type_c] = self._get_df(xml_tree.findall(element_type))
            # rename the columns (remove the namespaces)
            if element_type_c not in ns_dict[prf].keys():
                ns_dict[prf][element_type_c] = dict()
            for col in prf_content[element_type_c].columns:
                col_new = re.sub('[{].*?[}]', '', col)
                col_new = col_new.split('.')[-1]
                if col_new.endswith('-resource'):
                    col_new = col_new[:-9]
                    # remove the first character of each value if col_new is a CGMES class, e.g. Terminal
                    # otherwise remove the namespace from the literals (e.g. col_new is unitMultiplier, then the
                    # value is like http://iec.ch/TC57/2013/CIM-schema-cim16#UnitMultiplier.M
                    if col_new[0].isupper():
                        prf_content[element_type_c][col] = prf_content[element_type_c][col].str[1:]
                    elif prf_content[element_type_c][col].index.size > 0:
                        # get the namespace from the literal, Note: get the largest string because some values could
                        # be nan
                        name_space = \
                            prf_content[element_type_c][col].values[prf_content[element_type_c][col].str.len().idxmax()]
                        # remove the namespace from the literal
                        prf_content[element_type_c][col] = \
                            prf_content[element_type_c][col].str[name_space.rfind('.') + 1:]
                elif col_new.endswith('-about'):
                    col_new = 'rdfId'
                    prf_content[element_type_c][col] = prf_content[element_type_c][col].str[1:]
                elif col_new.endswith('-ID'):
                    col_new = 'rdfId'
                ns_dict[prf][element_type_c][col] = col_new
            prf_content[element_type_c] = prf_content[element_type_c].rename(columns={**ns_dict[prf][element_type_c]})
        if prf not in output.keys():
            output[prf] = prf_content
        else:
            for ele, df in prf_content.items():
                if ele not in output[prf].keys():
                    concat_list = [prf_content[ele]]
                else:
                    concat_list = [output[prf][ele], prf_content[ele]]
                output[prf][ele] = pd.concat(concat_list, ignore_index=True, sort=False)

    def _check_file(self, file: str) -> bool:
        if not os.path.isfile(file):
            self.logger.error("%s is not a valid file!" % file)
            self.report_container.add_log(Report(level=LogLevel.ERROR, code=ReportCode.ERROR_PARSING,
                                                 message="%s is not a valid file!" % file))
            return False
        elif file.lower().endswith('xml') or file.lower().endswith('rdf') or file.lower().endswith('zip'):
            return True
        else:
            return False

    def get_cim_dict(self) -> Dict[str, Dict[str, pd.DataFrame]]:
        return self.cim

    def set_cim_dict(self, cim: Dict[str, Dict[str, pd.DataFrame]]):
        self.cim = cim

    def get_file_names(self) -> Dict[str, str]:
        return self.file_names

    def get_report_container(self) -> ReportContainer:
        return self.report_container

    def _initialize_cim_data_structure(self, cgmes_version: str) -> Dict[str, Dict[str, pd.DataFrame]]:
        if cgmes_version == '2.4.15':
            return self._initialize_cim16_data_structure()
        if cgmes_version == '3.0':
            return self._initialize_cim100_data_structure()
        raise NotImplementedError(f"CGMES version {cgmes_version} is not supported.")

    def _initialize_cim100_data_structure(self) -> Dict[str, Dict[str, pd.DataFrame]]:
        """
           Get the cim data structure used by the converter for cgmes version 3.
           :return Dict[str, Dict[str, pd.DataFrame]]: The cim data structure used by the converter.
           """
        self.logger.debug("Returning the CIM data structure.")
        return MappingProxyType({
            'eq': MappingProxyType({
                'ControlArea': pd.DataFrame(columns=['rdfId', 'name', 'type']),
                'TieFlow': pd.DataFrame(columns=['rdfId', 'Terminal', 'ControlArea', 'positiveFlowIn']),
                'ConnectivityNode': pd.DataFrame(columns=['rdfId', 'name', 'description', 'ConnectivityNodeContainer']),
                'Bay': pd.DataFrame(columns=['rdfId', 'VoltageLevel']),
                'BusbarSection': pd.DataFrame(columns=['rdfId', 'name']),
                'Substation': pd.DataFrame(columns=['rdfId', 'name', 'Region']),
                'GeographicalRegion': pd.DataFrame(columns=['rdfId', 'name']),
                'SubGeographicalRegion': pd.DataFrame(columns=['rdfId', 'name', 'Region']),
                'VoltageLevel': pd.DataFrame(columns=['rdfId', 'name', 'shortName', 'BaseVoltage', 'Substation']),
                'BaseVoltage': pd.DataFrame(columns=['rdfId', 'name', 'nominalVoltage']),
                'ExternalNetworkInjection': pd.DataFrame(columns=[
                    'rdfId', 'name', 'description', 'minP', 'maxP', 'minQ', 'maxQ', 'BaseVoltage', 'EquipmentContainer',
                    'RegulatingControl', 'governorSCD']),
                'ACLineSegment': pd.DataFrame(columns=[
                    'rdfId', 'name', 'description', 'length', 'r', 'x', 'bch', 'gch', 'BaseVoltage', 'EquipmentContainer']),
                'Terminal': pd.DataFrame(columns=[
                    'rdfId', 'name', 'ConnectivityNode', 'ConductingEquipment', 'sequenceNumber']),
                'OperationalLimitSet': pd.DataFrame(columns=['rdfId', 'name', 'Terminal']),
                'OperationalLimitType': pd.DataFrame(columns=['rdfId', 'name', 'kind']),
                'CurrentLimit': pd.DataFrame(columns=[
                    'rdfId', 'name', 'OperationalLimitSet', 'OperationalLimitType']),
                'VoltageLimit': pd.DataFrame(columns=[
                    'rdfId', 'name', 'OperationalLimitSet', 'OperationalLimitType']),
                'DCNode': pd.DataFrame(columns=['rdfId', 'name', 'DCEquipmentContainer']),
                'DCEquipmentContainer': pd.DataFrame(columns=['rdfId', 'name']),
                'DCLine': pd.DataFrame(columns=['rdfId', 'name']),
                'DCConverterUnit': pd.DataFrame(columns=['rdfId', 'name', 'Substation', 'operationMode']),
                'DCLineSegment': pd.DataFrame(columns=['rdfId', 'name', 'description', 'EquipmentContainer']),
                'CsConverter': pd.DataFrame(columns=['rdfId', 'BaseVoltage', 'ratedUdc']),
                'VsConverter': pd.DataFrame(columns=['rdfId', 'name', 'BaseVoltage', 'EquipmentContainer', 'ratedUdc']),
                'DCTerminal': pd.DataFrame(columns=[
                    'rdfId', 'name', 'DCNode', 'DCConductingEquipment', 'sequenceNumber']),
                'ACDCConverterDCTerminal': pd.DataFrame(columns=[
                    'rdfId', 'name', 'DCNode', 'DCConductingEquipment', 'sequenceNumber']),
                'Breaker': pd.DataFrame(columns=[
                    'rdfId', 'name', 'description', 'EquipmentContainer', 'normalOpen', 'retained']),
                'Disconnector': pd.DataFrame(columns=[
                    'rdfId', 'name', 'description', 'EquipmentContainer', 'normalOpen', 'retained']),
                'Switch': pd.DataFrame(columns=[
                    'rdfId', 'name', 'description', 'EquipmentContainer', 'normalOpen', 'retained']),
                'LoadBreakSwitch': pd.DataFrame(columns=[
                    'rdfId', 'name', 'description', 'EquipmentContainer', 'normalOpen', 'retained']),
                'EnergyConsumer': pd.DataFrame(columns=[
                    'rdfId', 'name', 'description', 'BaseVoltage', 'EquipmentContainer']),
                'ConformLoad': pd.DataFrame(columns=['rdfId', 'name', 'description']),
                'NonConformLoad': pd.DataFrame(columns=['rdfId', 'name', 'description']),
                'StationSupply': pd.DataFrame(columns=['rdfId', 'name', 'description', 'BaseVoltage']),
                'GeneratingUnit': pd.DataFrame(columns=[
                    'rdfId', 'name', 'nominalP', 'minOperatingP', 'maxOperatingP', 'EquipmentContainer', 'governorSCD']),
                'WindGeneratingUnit': pd.DataFrame(columns=['rdfId', 'nominalP', 'minOperatingP', 'maxOperatingP', 'governorSCD']),
                'HydroGeneratingUnit': pd.DataFrame(columns=['rdfId', 'nominalP', 'minOperatingP', 'maxOperatingP', 'governorSCD']),
                'SolarGeneratingUnit': pd.DataFrame(columns=['rdfId', 'nominalP', 'minOperatingP', 'maxOperatingP', 'governorSCD']),
                'ThermalGeneratingUnit': pd.DataFrame(columns=['rdfId', 'nominalP', 'minOperatingP', 'maxOperatingP', 'governorSCD']),
                'NuclearGeneratingUnit': pd.DataFrame(columns=['rdfId', 'nominalP', 'minOperatingP', 'maxOperatingP', 'governorSCD']),
                'RegulatingControl': pd.DataFrame(columns=['rdfId', 'name', 'mode', 'Terminal']),
                'SynchronousMachine': pd.DataFrame(columns=[
                    'rdfId', 'name', 'description', 'GeneratingUnit', 'EquipmentContainer', 'ratedU', 'ratedS', 'type',
                    'InitialReactiveCapabilityCurve','ratedPowerFactor', 'minQ', 'maxQ', 'RegulatingControl']),
                'AsynchronousMachine': pd.DataFrame(columns=[
                    'rdfId', 'name', 'description', 'GeneratingUnit', 'ratedS', 'ratedU', 'ratedPowerFactor']),
                'EnergySource': pd.DataFrame(columns=[
                    'rdfId', 'name', 'description', 'nominalVoltage', 'EnergySchedulingType',
                    'EquipmentContainer']),
                'EnergySchedulingType': pd.DataFrame(columns=['rdfId', 'name']),
                'StaticVarCompensator': pd.DataFrame(columns=['rdfId', 'name', 'description', 'voltageSetPoint','sVCControlMode']),
                'PowerTransformer': pd.DataFrame(columns=['rdfId', 'name', 'description', 'EquipmentContainer']),
                'PowerTransformerEnd': pd.DataFrame(columns=[
                    'rdfId', 'name', 'PowerTransformer', 'endNumber', 'Terminal', 'ratedS', 'ratedU',
                    'r', 'x', 'b', 'g', 'BaseVoltage', 'connectionKind']),
                'TapChangerControl': pd.DataFrame(columns=['rdfId', 'name', 'mode', 'Terminal']),
                'RatioTapChanger': pd.DataFrame(columns=[
                    'rdfId', 'name', 'TransformerEnd', 'neutralStep', 'lowStep', 'highStep', 'stepVoltageIncrement',
                    'neutralU', 'normalStep', 'ltcFlag', 'TapChangerControl', 'RatioTapChangerTable']),
                'PhaseTapChangerLinear': pd.DataFrame(columns=[
                    'rdfId', 'TransformerEnd', 'neutralStep', 'lowStep', 'highStep', 'stepPhaseShiftIncrement',
                    'TapChangerControl']),
                'PhaseTapChangerAsymmetrical': pd.DataFrame(columns=[
                    'rdfId', 'TransformerEnd', 'neutralStep', 'lowStep', 'highStep', 'voltageStepIncrement',
                    'TapChangerControl']),
                'PhaseTapChangerSymmetrical': pd.DataFrame(columns=[
                    'rdfId', 'TransformerEnd', 'neutralStep', 'lowStep', 'highStep', 'voltageStepIncrement',
                    'TapChangerControl']),
                'PhaseTapChangerTabular': pd.DataFrame(columns=[
                    'rdfId', 'TransformerEnd', 'PhaseTapChangerTable', 'highStep', 'lowStep', 'neutralStep',
                    'TapChangerControl']),
                'PhaseTapChangerTablePoint': pd.DataFrame(columns=[
                    'rdfId', 'PhaseTapChangerTable', 'step', 'angle', 'ratio', 'r', 'x']),
                'RatioTapChangerTable': pd.DataFrame(columns=['rdfId', 'TransformerEnd', 'RatioTapChangerTable',
                                                              'highStep', 'lowStep', 'neutralStep']),
                'RatioTapChangerTablePoint': pd.DataFrame(columns=['rdfId', 'RatioTapChangerTable', 'step',
                                                                   'r', 'x', 'ratio']),
                'LinearShuntCompensator': pd.DataFrame(columns=[
                    'rdfId', 'name', 'description', 'nomU', 'gPerSection', 'bPerSection', 'maximumSections',
                    'normalSections']),
                'NonlinearShuntCompensator': pd.DataFrame(columns=[
                    'rdfId', 'name', 'description', 'nomU', 'maximumSections']),
                'NonlinearShuntCompensatorPoint': pd.DataFrame(columns=[
                    'rdfId', 'description', 'NonlinearShuntCompensator', 'sectionNumber', 'b', 'g']),
                'EquivalentBranch': pd.DataFrame(columns=[
                    'rdfId', 'name', 'description', 'BaseVoltage', 'r', 'x', 'r21', 'x21']),
                'EquivalentInjection': pd.DataFrame(columns=['rdfId', 'name', 'description', 'BaseVoltage']),
                'SeriesCompensator': pd.DataFrame(columns=[
                    'rdfId', 'name', 'description', 'BaseVoltage', 'r', 'x']),
                'MeasurementValueSource': pd.DataFrame(columns=['rdfId', 'name']),
                'PetersenCoil': pd.DataFrame(columns=['rdfId', 'name', 'description']),
                'ReactiveCapabilityCurve': pd.DataFrame(columns=['rdfId', 'name', 'curveStyle', 'xUnit', 'y1Unit',
                                                                 'y2Unit']),
                'CurveData': pd.DataFrame(columns=['rdfId', 'Curve', 'xvalue', 'y1value', 'y2value'])
            }),
            'eq_bd': MappingProxyType({
                'ConnectivityNode': pd.DataFrame(columns=['rdfId', 'name', 'ConnectivityNodeContainer']),
                'BaseVoltage': pd.DataFrame(columns=['rdfId', 'name', 'nominalVoltage']),
                'Terminal': pd.DataFrame(
                    columns=['rdfId', 'ConnectivityNode', 'ConductingEquipment', 'sequenceNumber']),
                'EnergySource': pd.DataFrame(columns=['rdfId', 'nominalVoltage']),
                'EnergySchedulingType': pd.DataFrame(columns=['rdfId', 'name'])
            }),
            'op': MappingProxyType({  # TODO: value attributes from AnalogValue and DiscreteValue need clarification
                'Analog': pd.DataFrame(columns=[
                    'rdfId', 'name', 'description', 'measurementType', 'unitSymbol', 'unitMultiplier', 'Terminal',
                    'PowerSystemResource', 'positiveFlowIn']),
                'AnalogValue': pd.DataFrame(columns=[
                    'rdfId', 'name', 'sensorAccuracy', 'MeasurementValueSource', 'Analog', 'value']),
                'Discrete': pd.DataFrame(columns=[
                    'rdfId', 'name', 'measurementType', 'unitSymbol', 'unitMultiplier', 'PowerSystemResource']),
                'DiscreteValue': pd.DataFrame(columns=[
                    'rdfId', 'name', 'sensorAccuracy', 'MeasurementValueSource', 'Discrete', 'value'])
            }),
            'sc': MappingProxyType({
                'PowerTransformer': pd.DataFrame(columns=['rdfId', 'isPartOfGeneratorUnit']),
                'PowerTransformerEnd': pd.DataFrame(columns=[
                    'rdfId', 'r0', 'x0', 'phaseAngleClock', 'grounded', 'xground']),
                'SynchronousMachine': pd.DataFrame(columns=['rdfId', 'r2', 'x2', 'voltageRegulationRange']),
                'AsynchronousMachine': pd.DataFrame(columns=[
                    'rdfId', 'rxLockedRotorRatio', 'iaIrRatio', 'efficiency', 'ratedMechanicalPower']),
                'ACLineSegment': pd.DataFrame(columns=[
                    'rdfId', 'r0', 'x0', 'b0ch', 'g0ch', 'shortCircuitEndTemperature']),
                'PetersenCoil': pd.DataFrame(columns=['rdfId', 'nominalU']),
                'ExternalNetworkInjection': pd.DataFrame(columns=[
                    'rdfId', 'maxInitialSymShCCurrent', 'minInitialSymShCCurrent', 'maxR1ToX1Ratio', 'minR1ToX1Ratio',
                    'maxR0ToX0Ratio', 'maxZ0ToZ1Ratio']),
                'EquivalentBranch': pd.DataFrame(columns=['rdfId', 'zeroR12', 'zeroR21', 'zeroX12', 'zeroX21']),
                'EquivalentInjection': pd.DataFrame(columns=['rdfId', 'r', 'x']),
                'SeriesCompensator': pd.DataFrame(columns=['rdfId', 'r0', 'x0']),
            }),
            'ssh': MappingProxyType({
                'ControlArea': pd.DataFrame(columns=['rdfId', 'netInterchange']),
                'ExternalNetworkInjection': pd.DataFrame(columns=[
                    'rdfId', 'p', 'q', 'referencePriority', 'controlEnabled', 'inService']),
                'Terminal': pd.DataFrame(columns=['rdfId', 'connected']),
                'DCTerminal': pd.DataFrame(columns=['rdfId', 'connected']),
                'ACDCConverterDCTerminal': pd.DataFrame(columns=['rdfId', 'connected']),
                'CsConverter': pd.DataFrame(columns=['rdfId', 'p', 'q', 'inService']),
                'VsConverter': pd.DataFrame(columns=[
                    'rdfId', 'p', 'q', 'targetUpcc', 'droop', 'droopCompensation', 'qShare', 'targetUdc', 'targetPpcc',
                    'targetQpcc', 'pPccControl', 'qPccControl', 'inService']),
                'Breaker': pd.DataFrame(columns=['rdfId', 'open', 'inService']),
                'Disconnector': pd.DataFrame(columns=['rdfId', 'open', 'inService']),
                'Switch': pd.DataFrame(columns=['rdfId', 'open', 'inService']),
                'LoadBreakSwitch': pd.DataFrame(columns=['rdfId', 'open', 'inService']),
                'EnergyConsumer': pd.DataFrame(columns=['rdfId', 'p', 'q', 'inService']),
                'ConformLoad': pd.DataFrame(columns=['rdfId', 'p', 'q', 'inService']),
                'NonConformLoad': pd.DataFrame(columns=['rdfId', 'p', 'q', 'inService']),
                'StationSupply': pd.DataFrame(columns=['rdfId', 'p', 'q', 'inService']),
                'RegulatingControl': pd.DataFrame(columns=[
                    'rdfId', 'discrete', 'enabled', 'targetValue', 'targetValueUnitMultiplier']),
                'SynchronousMachine': pd.DataFrame(columns=[
                    'rdfId', 'p', 'q', 'referencePriority', 'operatingMode', 'controlEnabled', 'inService']),
                'AsynchronousMachine': pd.DataFrame(columns=['rdfId', 'p', 'q', 'inService']),
                'EnergySource': pd.DataFrame(columns=[
                    'rdfId', 'activePower', 'reactivePower', 'inService', 'voltageAngle', 'voltageMagnitude']),
                'StaticVarCompensator': pd.DataFrame(columns=['rdfId', 'q', 'inService']),
                'TapChangerControl': pd.DataFrame(columns=[
                    'rdfId', 'discrete', 'enabled', 'targetValue', 'targetValueUnitMultiplier', 'targetDeadband']),
                'RatioTapChanger': pd.DataFrame(columns=['rdfId', 'step', 'controlEnabled']),
                'PhaseTapChangerLinear': pd.DataFrame(columns=['rdfId', 'step']),
                'PhaseTapChangerAsymmetrical': pd.DataFrame(columns=['rdfId', 'step']),
                'PhaseTapChangerSymmetrical': pd.DataFrame(columns=['rdfId', 'step']),
                'PhaseTapChangerTabular': pd.DataFrame(columns=['rdfId', 'step']),
                'LinearShuntCompensator': pd.DataFrame(columns=['rdfId', 'controlEnabled', 'sections', 'inService']),
                'NonlinearShuntCompensator': pd.DataFrame(columns=['rdfId', 'controlEnabled', 'sections', 'inService']),
                'EquivalentInjection': pd.DataFrame(columns=[
                    'rdfId', 'regulationTarget', 'regulationStatus', 'p', 'q', 'inService']),
                'GeneratingUnit': pd.DataFrame(columns=['rdfId', 'inService']),
                'NuclearGeneratingUnit': pd.DataFrame(columns=['rdfId', 'inService']),
                'HydroGeneratingUnit': pd.DataFrame(columns=['rdfId', 'inService']),
                'ThermalGeneratingUnit': pd.DataFrame(columns=['rdfId', 'inService']),
                'SolarGeneratingUnit': pd.DataFrame(columns=['rdfId', 'inService']),
                'WindGeneratingUnit': pd.DataFrame(columns=['rdfId', 'inService']),
                'CurrentLimit': pd.DataFrame(columns=['rdfId', 'value']),
                'VoltageLimit': pd.DataFrame(columns=['rdfId', 'value']),
            }),
            'sv': MappingProxyType({
                'SvVoltage': pd.DataFrame(columns=['rdfId', 'TopologicalNode', 'v', 'angle']),
                'SvPowerFlow': pd.DataFrame(columns=['rdfId', 'Terminal', 'p', 'q']),
                'SvShuntCompensatorSections': pd.DataFrame(columns=['rdfId', 'ShuntCompensator', 'sections']),
                'SvTapStep': pd.DataFrame(columns=['rdfId', 'TapChanger', 'position'])
            }),
            'tp': MappingProxyType({
                'TopologicalNode': pd.DataFrame(columns=[
                    'rdfId', 'name', 'description', 'ConnectivityNodeContainer', 'BaseVoltage']),
                'DCTopologicalNode': pd.DataFrame(columns=['rdfId', 'name', 'DCEquipmentContainer']),
                'ConnectivityNode': pd.DataFrame(columns=['rdfId', 'TopologicalNode']),
                'Terminal': pd.DataFrame(columns=['rdfId', 'TopologicalNode']),
                'DCTerminal': pd.DataFrame(columns=['rdfId', 'DCTopologicalNode']),
                'ACDCConverterDCTerminal': pd.DataFrame(columns=['rdfId', 'DCTopologicalNode'])
            }),
            'dl': MappingProxyType({
                'Diagram': pd.DataFrame(columns=['rdfId', 'name']),
                'DiagramObject': pd.DataFrame(columns=['rdfId', 'IdentifiedObject', 'Diagram', 'name']),
                'DiagramObjectPoint': pd.DataFrame(columns=[
                    'rdfId', 'sequenceNumber', 'xPosition', 'yPosition', 'DiagramObject'])}),
            'gl': MappingProxyType({
                'CoordinateSystem': pd.DataFrame(columns=['rdfId', 'name', 'crsUrn']),
                'Location': pd.DataFrame(columns=['rdfId', 'PowerSystemResources', 'CoordinateSystem']),
                'PositionPoint': pd.DataFrame(columns=['rdfId', 'Location', 'sequenceNumber', 'xPosition', 'yPosition'])
            })})

    def get_cim_data_structure(self) -> Dict[str, Dict[str, pd.DataFrame]]:
        cim_data_structure = {}
        for one_profile, one_profile_dict in self.__cim_blueprint.items():
            cim_data_structure[one_profile] = {}
            for one_class, one_class_df in one_profile_dict.items():
                cim_data_structure[one_profile][one_class] = one_class_df.copy(deep=True)
        return cim_data_structure



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2023 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.


import warnings

import numpy as np
from scipy.sparse.linalg import inv as inv_sparse
from scipy.linalg import inv

from pandapower.pypower.idx_bus_sc import R_EQUIV, X_EQUIV
from pandapower.pypower.idx_bus import BASE_KV
from pandapower.auxiliary import _clean_up

try:
    from pandapower.pf.makeYbus_numba import makeYbus
except ImportError:
    from pandapower.pypower.makeYbus import makeYbus


def _calc_rx(net, ppci, bus_idx):
    # Vectorized for multiple bus
    r_fault = net["_options"]["r_fault_ohm"]
    x_fault = net["_options"]["x_fault_ohm"]
    if r_fault > 0 or x_fault > 0:
        base_r = np.square(ppci["bus"][bus_idx, BASE_KV]) / ppci["baseMVA"]
        fault_impedance = (r_fault + x_fault * 1j) / base_r
    else:
        fault_impedance = 0 + 0j
    net._options["fault_impedance"] = fault_impedance

    if net["_options"]["inverse_y"]:
        Zbus = ppci["internal"]["Zbus"]
        z_equiv = np.diag(Zbus)[bus_idx] + fault_impedance
    else:
        z_equiv = _calc_zbus_diag(net, ppci, bus_idx) + fault_impedance
    ppci["bus"][bus_idx, R_EQUIV] = z_equiv.real
    ppci["bus"][bus_idx, X_EQUIV] = z_equiv.imag


def _calc_ybus(ppci):
    Ybus, Yf, Yt = makeYbus(ppci["baseMVA"], ppci["bus"], ppci["branch"])
    if np.isnan(Ybus.data).any():
        raise ValueError("nan value detected in Ybus matrix - check calculation parameters for nan values")
    ppci["internal"]["Yf"] = Yf
    ppci["internal"]["Yt"] = Yt
    ppci["internal"]["Ybus"] = Ybus


def _calc_zbus(net, ppci):
    try:
        Ybus = ppci["internal"]["Ybus"]
        sparsity = Ybus.nnz / Ybus.shape[0]**2
        if sparsity < 0.002:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                ppci["internal"]["Zbus"] = inv_sparse(Ybus).toarray()
        else:
            ppci["internal"]["Zbus"] = inv(Ybus.toarray())
    except Exception as e:
        _clean_up(net, res=False)
        raise (e)


def _calc_zbus_diag(net, ppci, bus_idx=None):
    ybus_fact = ppci["internal"]["ybus_fact"]
    n_ppci_bus = ppci["bus"].shape[0]

    if bus_idx is None:
        diagZ = np.zeros(n_ppci_bus, dtype=np.complex128)
        for i in range(n_ppci_bus):
            b = np.zeros(n_ppci_bus, dtype=np.complex128)
            b[i] = 1 + 0j
            diagZ[i] = ybus_fact(b)[i]
        ppci["internal"]["diagZ"] = diagZ
        return diagZ
    else:
        diagZ = np.zeros(bus_idx.shape[0], dtype=np.complex128)
        for ix, b in enumerate(bus_idx):
            rhs = np.zeros(n_ppci_bus, dtype=np.complex128)
            rhs[b] = 1 + 0j
            diagZ[ix] = ybus_fact(rhs)[b]
        return diagZ



import numpy as np
from pandapower.control.basic_controller import Controller

try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging

class ShuntController(Controller):
    """
    Base Shunt Controller for controlling the steps of a shunt in a power network.

    The `ShuntController` provides a basic framework for controlling a shunt in a pandapower network. It adjusts
    the reactive power compensation provided by a shunt in discrete steps to influence the voltage level at a
    particular bus. This class is intended to be extended by more specific controller implementations, such as
    the `DiscreteShuntController`.

    INPUT:
        **net** (attrdict) - pandapower network object

        **shunt_index** (int) - The index of the shunt in the pandapower network that is being controlled.

    OPTIONAL:
        **bus_index** (int, None) - The index of the bus that the controller monitors and regulates the voltage at.
                                    If None, the bus connected to the shunt is used by default.

        **tol** (float, 0.001) - Voltage tolerance in per-unit (pu) for controlling the bus voltage. This defines
                                 the acceptable range around the setpoint where no control action is taken.

        **in_service** (bool, True) - Boolean flag indicating whether the controller is currently active or not.

        **check_step_bounds** (bool, True) - Flag to check if the shunt steps should be constrained within predefined
                                             minimum and maximum step limits.

        **order** (int, 0) - The execution order of the controller, which determines when this controller is executed
                             relative to others in the simulation.

        **level** (int, 0) - The hierarchy level of the controller in multi-level control schemes, allowing control
                             actions to be prioritized.

        **kwargs** - Additional keyword arguments passed to the base Controller class.
    """
    def __init__(self, net, shunt_index, bus_index=None, tol=1e-3, in_service=True,
                 check_step_bounds=True, order=0, level=0, **kwargs):
        super().__init__(net, in_service=in_service, order=order, level=level,
                         **kwargs)
        self.tol = tol
        self.shunt_index = shunt_index
        self.element_in_service = net.shunt.loc[self.shunt_index, 'in_service']
        if bus_index is None:
            self.controlled_bus = net.shunt.at[self.shunt_index, 'bus']
        else:
            self.controlled_bus = bus_index

        self.step = net.shunt.at[shunt_index, 'step']

        self.check_step_bounds = check_step_bounds
        if check_step_bounds:
            self.step_min = 0
            self.step_max = net.shunt.at[self.shunt_index, 'max_step']

        ext_grid_buses = net.ext_grid.loc[net.ext_grid.in_service, 'bus'].values
        if self.controlled_bus in ext_grid_buses:
            logging.warning("Controlled Bus is Slack Bus - deactivating controller")
            self.set_active(net, False)

class DiscreteShuntController(ShuntController):
    """
    Discrete Shunt Controller that controls the shunt steps in a discrete manner to regulate the voltage
    at a specific bus. This controller adjusts the active and reactive power of a shunt in steps based on the
    voltage deviation from a setpoint.

    INPUT:
        **net** (attrdict) - pandapower network object

        **shunt_index** (int) - The index of the shunt in the pandapower network to be controlled.

        **vm_set_pu** (float) - The voltage setpoint in per-unit (pu) for controlling the bus voltage.

    OPTIONAL:
        **bus_index** (int, None) - The index of the bus where voltage control is applied. If None, the bus
                                    connected to the shunt is used by default.

        **tol** (float, 0.001) - Voltage tolerance band in per-unit (pu) for control action (default is 1% or 0.01 pu).

        **increment** (int, 1) - Step increment size for controlling the shunt. The controller adjusts the
                                 shunt steps by this increment based on the voltage deviation.

        **reset_at_init** (bool, False) - If True, the shunt steps will be reset to 0 during the initialization
                                          of the controller.

        **in_service** (bool, True) - Boolean flag to indicate whether the controller is active or not.

        **check_step_bounds** (bool, True) - If True, the controller will check and enforce the step boundaries
                                             (minimum and maximum) for the shunt.

        **order** (int, 0) - Execution order of the controller. Controllers with lower order are executed first.

        **level** (int, 0) - Controller level.

        **matching_params** (dict, None) - Dictionary of parameters used to match this controller with the appropriate
                                           elements in the network. Defaults to shunt_index and bus_index.

    """
    def __init__(self, net, shunt_index, vm_set_pu, bus_index=None, tol=1e-3, increment=1, reset_at_init=False,
                 in_service=True, check_step_bounds=True, order=0, level=0, matching_params=None, **kwargs):
        if matching_params is None:
            matching_params = {"shunt_index": shunt_index, "bus_index": bus_index}
        super().__init__(net, shunt_index=shunt_index, bus_index=bus_index, tol=tol,
                         in_service=in_service,
                         check_step_bounds=check_step_bounds, order=order, level=level,
                         matching_params=matching_params, **kwargs)
        self.reset_at_init = reset_at_init

        self.vm_set_pu = vm_set_pu
        self.step = net.shunt.at[self.shunt_index, 'step']
        self.increment = increment
        if not check_step_bounds:
            net.shunt.step = net.shunt.step.astype(np.int64)

    def initialize_control(self, net):
        if self.reset_at_init:
            self.step = 0
            net.shunt.at[self.shunt_index, 'step'] = 0

    def control_step(self, net):
        vm_pu = net.res_bus.at[self.controlled_bus, 'vm_pu']
        self.step = net.shunt.at[self.shunt_index, "step"]

        sign = np.sign(net.shunt.at[self.shunt_index, 'q_mvar'])
        if vm_pu > self.vm_set_pu + self.tol:
            self.step += self.increment * sign
        elif vm_pu <= self.vm_set_pu - self.tol:
            self.step -= self.increment * sign

        if self.check_step_bounds:
            self.step = np.clip(self.step, self.step_min, self.step_max)

        # Write to net
        net.shunt.at[self.shunt_index, 'step'] = self.step

    def is_converged(self, net):
        if not net.shunt.at[self.shunt_index, 'in_service']:
            return True

        vm_pu = net.res_bus.at[self.controlled_bus, "vm_pu"]
        if abs(vm_pu - self.vm_set_pu) < self.tol:
            return True

        if self.check_step_bounds:
            if net.shunt.at[self.shunt_index, 'q_mvar'] >= 0:
                if vm_pu < self.vm_set_pu and self.step == self.step_min:
                    return True
                elif vm_pu > self.vm_set_pu and self.step == self.step_max:
                    return True
            else:
                if vm_pu < self.vm_set_pu and self.step == self.step_max:
                    return True
                elif vm_pu > self.vm_set_pu and self.step == self.step_min:
                    return True
        return False